# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def npu_group_norm_silu(x, gama, beta, group, eps):
    if len(x.shape) != 4:
        raise ValueError("x shape should be (N, C, H, W)")
    res = torch.ops.aten.native_group_norm(x, gama, beta, x.shape[0], x.shape[1], x.shape[2] * x.shape[3], group, eps)
    res = list(res)
    if not res:
        raise ValueError("run native_group_norm failed")
    res[0] = torch.nn.functional.silu(res[0])
    return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

# 进行比对及结果展示
import os
from collections import namedtuple

import numpy as np
from msprobe.core.common.utils import CompareException
from msprobe.core.common.file_utils import get_json_contents, write_csv
import torch
from msprobe.core.common.const import CompareConst
from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_register import StandardRegistry
from msprobe.pytorch.api_accuracy_checker.precision_standard.absolute_threshold import AbsolutethdCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.benchmark_compare import BenchmarkCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.ulp_compare import UlpCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.binary_consistency import BinaryCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.thousandth_standard import ThousandthStdCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.accumulative_error_compare import AccumulativeErrorCompare
from msprobe.pytorch.api_accuracy_checker.compare.compare_input import CompareInput
from msprobe.pytorch.api_accuracy_checker.compare.algorithm import get_abs_err, get_max_abs_err, get_rel_err_ratio, \
    cosine_sim, get_rel_err_origin, get_abs_bench_with_eps, compare_bool_tensor
from msprobe.pytorch.api_accuracy_checker.common.config import msCheckerConfig
from msprobe.pytorch.api_accuracy_checker.compare.compare_column import CompareColumn
from msprobe.pytorch.api_accuracy_checker.compare.compare_utils import check_dtype_comparable, \
    DETAIL_TEST_ROWS, BENCHMARK_COMPARE_SUPPORT_LIST
from msprobe.pytorch.api_accuracy_checker.common.utils import extract_basic_api_segments
from msprobe.pytorch.common.log import logger
from msprobe.core.common.decorator import recursion_depth_decorator


ResultInfo = namedtuple('ResultInfo', ['full_api_name', 'fwd_success_status', 'bwd_success_status',
                                       'fwd_compare_alg_results', 'bwd_compare_alg_results', 'rank'])


INDEX_TEST_RESULT_GROUP = 3
BACKWARD_RESULT_GROUP = 4
INDEX_FIRST_GROUP = 0
INDEX_MESSAGE = -1


class Comparator:
    # consts for result csv
    COLUMN_API_NAME = "API name"
    COLUMN_FORWARD_SUCCESS = "Forward Test Success"
    COLUMN_BACKWARD_SUCCESS = "Backward Test Success"
    COLUMN_STACK_INFO = "Traceback callstack info"

    def __init__(self, result_csv_path, details_csv_path, is_continue_acc_check,
                 stack_info_json_path=None, config=None):
        self.save_path_str = result_csv_path
        self.detail_save_path_str = details_csv_path
        self.save_path_list = [result_csv_path]
        self.detail_save_path_list = [details_csv_path]

        self.registry = self._register_compare_func()

        if not is_continue_acc_check:
            self.write_csv_title()
        if stack_info_json_path:
            self.stack_info = get_json_contents(stack_info_json_path)
        else:
            self.stack_info = None

    @staticmethod
    def get_path_from_rank(rank, path_list, path_pattern):
        return path_list[-1] if len(path_list) == 1 else path_pattern.format(rank)

    @staticmethod
    def print_pretest_result():
        logger.info("Successfully completed acc_check/multi_acc_check.")

    @staticmethod
    def _compare_dropout(bench_output, device_output):
        tensor_num = bench_output.numel()
        if tensor_num >= 100:
            if abs((bench_output == 0).sum() - (device_output == 0).cpu().sum()) / tensor_num < 0.1:
                return CompareConst.PASS, 1
            else:
                return CompareConst.ERROR, 0
        else:
            return CompareConst.PASS, 1

    @staticmethod
    def _compare_builtin_type(bench_output, device_output, compare_column):
        if not isinstance(bench_output, (bool, int, float, str)):
            return CompareConst.PASS, compare_column, ""
        if bench_output != device_output:
            return CompareConst.ERROR, compare_column, ""
        compare_column.error_rate = 0
        return CompareConst.PASS, compare_column, ""

    @staticmethod
    def _get_acc_check_detail(test_result):
        """get acc_check detail before write to csv, called by online acc_check"""
        test_rows = []
        try:
            subject_prefix = test_result[0]
            fwd_result = test_result[3]
            bwd_result = test_result[4]
        except IndexError as e:
            logger.error("List index out of bounds when writing detail CSV.")
            raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR, "list index out of bounds") from e

        if isinstance(fwd_result, list):
            for i, test_subject in enumerate(fwd_result):
                subject = subject_prefix + ".forward.output." + str(i)
                test_subject = ["{:.{}f}".format(item, msCheckerConfig.precision)
                                if isinstance(item, float) else item for item in test_subject]
                test_rows.append([subject] + list(test_subject))
        if isinstance(bwd_result, list):
            for i, test_subject in enumerate(bwd_result):
                subject = subject_prefix + ".backward.output." + str(i)
                test_subject = ["{:.{}f}".format(item, msCheckerConfig.precision)
                                if isinstance(item, float) else item for item in test_subject]
                test_rows.append([subject] + list(test_subject))
        return test_rows

    @staticmethod
    def _binary_standard_compare(input_data):
        binary_compare = BinaryCompare(input_data)
        binary_compare.compare()
    
    @staticmethod
    def _thousandth_standard_compare(input_data):
        thousandth_compare = ThousandthStdCompare(input_data)
        thousandth_compare.compare()
    
    @staticmethod
    def _absolute_standard_compare(input_data):
        absolute_compare = AbsolutethdCompare(input_data)
        absolute_compare.compare()

    @staticmethod
    def _ulp_compare(input_data):
        ulp_compare = UlpCompare(input_data)
        ulp_compare.compare()
    
    @staticmethod
    def _benchmark_compare(input_data):
        benchmark_compare = BenchmarkCompare(input_data)
        benchmark_compare.compare()
    
    @staticmethod
    def _accumulative_error_compare(input_data):
        accumulative_error_compare = AccumulativeErrorCompare(input_data)
        accumulative_error_compare.compare()

    def write_csv_title(self):
        summary_test_rows = [
            [self.COLUMN_API_NAME, 
             self.COLUMN_FORWARD_SUCCESS,
             self.COLUMN_BACKWARD_SUCCESS,
             "Message"]
            ]
        for save_path, detail_save_path in zip(self.save_path_list, self.detail_save_path_list):
            if not os.path.exists(save_path):
                write_csv(summary_test_rows, save_path)
            if not os.path.exists(detail_save_path):
                write_csv(DETAIL_TEST_ROWS, detail_save_path)

    @recursion_depth_decorator("compare_core")
    def _compare_core(self, api_name, bench_output, device_output, is_fp8):
        compare_column = CompareColumn()
        if not isinstance(bench_output, type(device_output)):
            status = CompareConst.ERROR
            message = "bench and npu output type is different."
        elif isinstance(bench_output, dict):
            b_keys, n_keys = set(bench_output.keys()), set(device_output.keys())
            if b_keys != n_keys:
                status = CompareConst.ERROR
                message = "bench and npu output dict keys are different."
            else:
                status, compare_column, message = self._compare_core(api_name, list(bench_output.values()),
                                                                     list(device_output.values()), is_fp8)
        elif isinstance(bench_output, torch.Tensor):
            copy_bench_out = bench_output.detach().clone()
            copy_device_output = device_output.detach().clone()
            compare_column.bench_type = str(copy_bench_out.dtype)
            compare_column.npu_type = str(copy_device_output.dtype)
            compare_column.shape = tuple(device_output.shape)
            status, compare_column, message = self._compare_torch_tensor(api_name, copy_bench_out, copy_device_output,
                                                                         compare_column, is_fp8)
        elif isinstance(bench_output, (bool, int, float, str)):
            compare_column.bench_type = str(type(bench_output))
            compare_column.npu_type = str(type(device_output))
            status, compare_column, message = self._compare_builtin_type(bench_output, device_output, compare_column)
        elif bench_output is None:
            status = CompareConst.SKIP
            message = "Bench output is None, skip this test."
        else:
            status = CompareConst.ERROR
            message = "Unexpected output type in compare_core: {}".format(type(bench_output))

        return status, compare_column, message
    
    def write_summary_csv(self, test_result):
        test_rows = []
        try:
            name = test_result[0]
            df_row = list(test_result[:INDEX_TEST_RESULT_GROUP])
            if test_result[1] == CompareConst.SKIP:
                df_row.append(test_result[INDEX_TEST_RESULT_GROUP][INDEX_FIRST_GROUP][INDEX_MESSAGE])
            elif test_result[2] == CompareConst.SKIP:
                df_row.append(test_result[BACKWARD_RESULT_GROUP][INDEX_FIRST_GROUP][INDEX_MESSAGE])
            if self.stack_info:
                stack_info = "\n".join(self.stack_info[name])
                df_row.append(stack_info)
            test_rows.append(df_row)
            save_path = self.get_path_from_rank(test_result[-1], self.save_path_list, self.save_path_str)
        except IndexError as e:
            logger.error("List index out of bounds when writing summary CSV.")
            raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR, "list index out of bounds") from e
        write_csv(test_rows, save_path)

    def write_detail_csv(self, test_result):
        test_rows = self._get_acc_check_detail(test_result)
        detail_save_path = self.get_path_from_rank(test_result[-1],
                                                   self.detail_save_path_list,
                                                   self.detail_save_path_str)
        write_csv(test_rows, detail_save_path)

    def record_results(self, args):
        self.write_summary_csv(args)
        self.write_detail_csv(args)


    def compare_output(self, full_api_name, data_info):
        """Get compare result and write to result and detail csv.
        """
        _, api_name = extract_basic_api_segments(full_api_name)
        if not api_name:
            raise ValueError(f"API name {full_api_name} has not been adapted.")
        bench_output, device_output = data_info.bench_output, data_info.device_output
        bench_grad, device_grad = data_info.bench_grad, data_info.device_grad
        backward_message = data_info.backward_message
        is_fp8 = data_info.is_fp8
        if "dropout" in full_api_name:
            fwd_success_status, fwd_compare_alg_results = self._compare_dropout(bench_output, device_output)
        else:
            fwd_success_status, fwd_compare_alg_results = self._compare_core_wrapper(api_name, bench_output,
                                                                                     device_output, is_fp8)
        if not (bench_grad and device_grad):
            bwd_success_status, bwd_compare_alg_results = (CompareConst.SPACE, [])
        else:
            if "dropout" in full_api_name:
                bwd_success_status, bwd_compare_alg_results = self._compare_dropout(bench_grad[0], device_grad[0])
            else:
                bwd_success_status, bwd_compare_alg_results = self._compare_core_wrapper(api_name, bench_grad,
                                                                                         device_grad, is_fp8)
        if backward_message:
            backward_column = CompareColumn()
            bwd_compare_alg_results = [backward_column.to_column_value(CompareConst.SKIP, backward_message)]
            bwd_success_status = CompareConst.SKIP
        else:
            bwd_success_status = bwd_success_status if bwd_compare_alg_results is not None else CompareConst.SPACE
        result_info = ResultInfo(full_api_name,
                                 fwd_success_status,
                                 bwd_success_status,
                                 fwd_compare_alg_results,
                                 bwd_compare_alg_results,
                                 data_info.rank)
        self.record_results(result_info)
        return fwd_success_status == CompareConst.PASS, bwd_success_status == CompareConst.PASS \
               or bwd_success_status == CompareConst.SPACE

    def _register_compare_func(self):
        registry = StandardRegistry()
        registry.register(CompareConst.ABSOLUTE_THRESHOLD, self._absolute_standard_compare)
        registry.register(CompareConst.BINARY_CONSISTENCY, self._binary_standard_compare)
        registry.register(CompareConst.ULP_COMPARE, self._ulp_compare)
        registry.register(CompareConst.THOUSANDTH_STANDARD, self._thousandth_standard_compare)
        registry.register(CompareConst.BENCHMARK, self._benchmark_compare)
        registry.register(CompareConst.ACCUMULATIVE_ERROR_COMPARE, self._accumulative_error_compare)
        return registry

    def _compare_core_wrapper(self, api_name, bench_output, device_output, is_fp8):
        detailed_result_total = []
        test_final_success = CompareConst.PASS
        if isinstance(bench_output, (list, tuple)):
            status, compare_result, message = [], [], []
            if len(bench_output) > len(device_output):
                status = [CompareConst.ERROR]
                message = ["bench and npu output structure is different."]
            else:
                device_output = device_output[:len(bench_output)]
                for b_out_i, n_out_i in zip(bench_output, device_output):
                    status_i, compare_result_i, message_i = self._compare_core(api_name, b_out_i, n_out_i, is_fp8)
                    status.append(status_i)
                    compare_result.append(compare_result_i)
                    message.append(message_i)
        else:
            status, compare_result, message = self._compare_core(api_name, bench_output, device_output, is_fp8)
        if not isinstance(status, list):
            detailed_result_total.append(compare_result.to_column_value(status, message))
            if status == CompareConst.ERROR:
                test_final_success = CompareConst.ERROR
            elif status == CompareConst.WARNING:
                test_final_success = CompareConst.WARNING
        else:
            for item, item_status in enumerate(status):
                detailed_result_total.append(compare_result[item].to_column_value(item_status, message[item]))
                if item_status == CompareConst.ERROR:
                    test_final_success = CompareConst.ERROR
                elif item_status == CompareConst.WARNING:
                    test_final_success = CompareConst.WARNING
        return test_final_success, detailed_result_total

    def _compare_torch_tensor(self, api_name, bench_output, device_output, compare_column, is_fp8):
        cpu_shape = bench_output.shape
        npu_shape = device_output.shape
        
        npu_dtype = device_output.dtype
        if is_fp8:
            in_dtype = torch.float8_e4m3fn
        else:
            in_dtype = torch.float32
        if npu_dtype == torch.bfloat16:
            bench_output = bench_output.to(torch.float32)
            device_output = device_output.to(torch.float32)
        bench_output = bench_output.cpu().numpy()
        device_output = device_output.cpu().numpy()
        if cpu_shape != npu_shape:
            return CompareConst.ERROR, compare_column, f"The shape of bench{str(cpu_shape)} " \
                                                       f"and npu{str(npu_shape)} not equal."
        if not check_dtype_comparable(bench_output, device_output):
            return CompareConst.ERROR, compare_column, f"Bench out dtype is {bench_output.dtype} but " \
                                                       f"npu output dtype is {device_output.dtype}, cannot compare."
        message = ""
        if bench_output.size == 0:
            return CompareConst.ERROR, compare_column, "There is not bench calculation result."
        if bench_output.dtype in [bool, np.uint8, np.int8, np.int16, np.uint16, np.uint32, np.int32,
                                  np.int64, np.uint64]:
            message += f"Compare algorithm is not supported for {bench_output.dtype} data. " \
                       f"Only judged by Error Rate."
            err_rate, status, msg = compare_bool_tensor(bench_output, device_output)
            message += msg + "\n"
            compare_column.error_rate = err_rate
            return status, compare_column, message
        else:
            in_and_out_dtype = {
                'dtype': npu_dtype,
                'in_dtype': in_dtype
            }
            status, compare_column, message = self._compare_float_tensor(api_name, bench_output, device_output,
                                                                         compare_column, in_and_out_dtype)
            return status, compare_column, message

    def _perform_comparison(self, api_name, input_data, dtype, in_dtype):
        comparison_func = self.registry.get_comparison_function(api_name, dtype, in_dtype)
        comparison_func(input_data)
            
    def _compare_float_tensor(self, api_name, bench_output, device_output, compare_column, in_and_out_dtype):
        dtype = in_and_out_dtype.get('dtype')
        in_dtype = in_and_out_dtype.get('in_dtype')
        message = ""
        _, abs_bench_with_eps = get_abs_bench_with_eps(bench_output, dtype)
        abs_err = get_abs_err(bench_output, device_output)
        rel_err_orign = get_rel_err_origin(abs_err, abs_bench_with_eps)
        input_data = CompareInput(bench_output, device_output, compare_column, dtype, rel_err_orign)
        if str(dtype) in BENCHMARK_COMPARE_SUPPORT_LIST:
            self._perform_comparison(api_name, input_data, dtype, in_dtype)
        else:
            message += f"The data type {dtype} is not supported for new precision standard."

        cos_res, cos_status, msg = cosine_sim(bench_output, device_output)
        compare_column.cosine_sim = cos_res
        message += msg + "\n"
        if not cos_status:
            message += "Cosine similarity is less than 0.99, consider as error, skip other check and set to SPACE.\n"
            return CompareConst.ERROR, compare_column, message

        max_abs_res, max_abs_status = get_max_abs_err(abs_err)
        compare_column.max_abs_err = max_abs_res
        if max_abs_status:
            message += "Max abs error is less than 0.001, consider as pass, skip other check and set to SPACE.\n"
            return CompareConst.PASS, compare_column, message

        if dtype in [torch.float16, torch.bfloat16]:
            hundred_res, hundred_status = get_rel_err_ratio(rel_err_orign, CompareConst.HUNDRED_RATIO_THRESHOLD)
            compare_column.rel_err_hundredth = hundred_res
            if not hundred_status:
                message += "Relative error is greater than 0.01, consider as error, " \
                           "skip other check and set to SPACE.\n"
                return CompareConst.ERROR, compare_column, message
        thousand_res, thousand_status = get_rel_err_ratio(rel_err_orign, CompareConst.THOUSAND_RATIO_THRESHOLD)
        compare_column.rel_err_thousandth = thousand_res
        if dtype in [torch.float16, torch.bfloat16]:
            if thousand_status:
                message += "Relative error is less than 0.001, consider as pass, skip other check and set to SPACE.\n"
                return CompareConst.PASS, compare_column, message
            message += "Relative error is greater than 0.001, consider as warning, skip other check and set to SPACE.\n"
            return CompareConst.WARNING, compare_column, message
        ten_thousand_res, ten_thousand_status = get_rel_err_ratio(
                                                rel_err_orign, CompareConst.TEN_THOUSAND_RATIO_THRESHOLD)
        compare_column.rel_err_ten_thousandth = ten_thousand_res
        if dtype in [torch.float32, torch.float64]:
            if not thousand_status:
                message += "Relative error is greater than 0.001, consider as error, " \
                           "skip other check and set to SPACE.\n"
                return CompareConst.ERROR, compare_column, message
            if not ten_thousand_status:
                message += "Relative error is greater than 0.0001, consider as warning, " \
                           "skip other check and set to SPACE.\n"
                return CompareConst.WARNING, compare_column, message
            message += "Relative error is less than 0.0001, consider as pass.\n"
        return CompareConst.PASS, compare_column, message



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
convert format from FRACTAL_Z to HWCN.
"""
import numpy as np


def convert(shape_from: list, shape_to: list, array: any) -> any:
    """
    Convert the data format from FRACTAL_Z to HWCN
    :param shape_from: the shape before convert
    :param shape_to: the shape after convert
    :param array: the one-dimensional array
    :return: the data array of HWCN shape
    """
    axis_c = shape_to[2]
    axis_n = shape_to[3]
    axis_no = shape_from[1]
    axis_ni = shape_from[2]
    axis_h = shape_to[0]
    axis_w = shape_to[1]
    axis_c1 = shape_from[0] // (axis_h * axis_w)
    axis_c0 = shape_from[3]
    c_pad = None if axis_c1 * axis_c0 == axis_c else axis_c - axis_c1 * axis_c0
    n_pad = None if axis_no * axis_ni == axis_n else axis_n - axis_no * axis_ni
    tmp_input_tensor = array.reshape(axis_c1, axis_h, axis_w, axis_no, axis_ni, axis_c0)
    # transpose the shape from (c1,h,w,no,ni,c0) to (h,w,c1,c0,no,ni)
    tmp_input_tensor = np.transpose(tmp_input_tensor, (1, 2, 0, 5, 3, 4))
    tmp_input_tensor = tmp_input_tensor.reshape((axis_h, axis_w, axis_c1 * axis_c0, axis_no * axis_ni))
    return tmp_input_tensor[:, :, :c_pad, :n_pad]



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import os
import json
import pandas as pd

from msprobe.core.common.file_utils import create_file_in_zip, os_walk_for_files, load_json
from msprobe.core.config_check.checkers.base_checker import BaseChecker
from msprobe.core.config_check.config_checker import register_checker_item, register_pre_forward_fun_list
from msprobe.core.config_check.utils.utils import config_checking_print, get_tensor_features
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.const import Const


def collect_weights_data(model):
    weights_data = {}
    for name, param in FmkAdp.named_parameters(model):
        if param.dtype != FmkAdp.dtype("float32"):
            param = param.float()
        weights_data[name] = get_tensor_features(param)
    return weights_data


def compare_weight_file(bench_file, cmp_file):
    bench_data = load_json(bench_file)
    cmp_data = load_json(cmp_file)

    results = []
    for weight_name in set(bench_data.keys()) | set(cmp_data.keys()):
        result = {
            "weight_name": weight_name,
            "equal": None,
            "max_relative_diff": None,
            "min_relative_diff": None,
            "mean_relative_diff": None,
            "norm_relative_diff": None
        }

        if weight_name not in bench_data:
            result["equal"] = "only cmp have"
            results.append(result)
            continue

        if weight_name not in cmp_data:
            result["equal"] = "only bench have"
            results.append(result)
            continue

        bench_vals = bench_data[weight_name]
        cmp_vals = cmp_data[weight_name]
        keys = ["max", "min", "mean", "norm"]
        equal = all([bench_vals[k] == cmp_vals[k] for k in keys])
        result["equal"] = equal

        for key in keys:
            diff_key = f"{key}_relative_diff"
            result[diff_key] = (abs(bench_vals[key] - cmp_vals[key]) / bench_vals[key]) \
            if bench_vals[key] != 0 else None

        results.append(result)

    return results


def compare_weight(bench_dir, cmp_dir):
    all_results = []
    bench_files_info = os_walk_for_files(bench_dir, 10)
    for info in bench_files_info:
        if not info["file"].endswith('.json'):
            continue
        bench_file = os.path.join(info["root"], info["file"])
        relative_path = os.path.relpath(info["root"], bench_dir)
        cmp_root = os.path.join(cmp_dir, relative_path)
        cmp_file = os.path.join(cmp_root, info["file"])

        path_list = relative_path.split(os.sep)
        if len(path_list) < 2:
            raise Exception("Can not compare weights because the extracted file has been corrupted!")
        step = int(path_list[0].replace("step", ""))
        rank = int(path_list[1].replace("rank", ""))

        if not os.path.exists(cmp_file):
            bench_data = load_json(bench_file)
            for weight_name in bench_data.keys():
                result = {
                    "step": step,
                    "rank": rank,
                    "weight_name": weight_name,
                    "equal": "only bench have",
                    "max_relative_diff": None,
                    "min_relative_diff": None,
                    "mean_relative_diff": None,
                    "norm_relative_diff": None
                }
                all_results.append(result)
        else:
            results = compare_weight_file(bench_file, cmp_file)
            for res in results:
                res["step"] = step
                res["rank"] = rank
                all_results.append(res)

    df = pd.DataFrame(all_results, columns=WeightsChecker.result_header)
    df = df.sort_values(by=['step', 'rank'], ascending=[True, True])
    return df


@register_checker_item("weights")
class WeightsChecker(BaseChecker):
    input_needed = "model"
    multi_rank = True

    target_name_in_zip = "weights"
    result_header = ["step", "rank", "weight_name", "equal", "max_relative_diff", 
                     "min_relative_diff", "mean_relative_diff", "norm_relative_diff"]

    @staticmethod
    def pack(pack_input):
        output_zip_path = pack_input.output_zip_path

        def collect_weights(model, args, kwargs, step):
            weights_data_dict = collect_weights_data(model)
            weights_data_filepath = os.path.join(WeightsChecker.target_name_in_zip, 
                                                 f"step{step}", f"rank{FmkAdp.get_rank_id()}", "weight.json")
            create_file_in_zip(output_zip_path, weights_data_filepath, json.dumps(weights_data_dict, indent=4))
            config_checking_print(f"add weights info to zip")
        register_pre_forward_fun_list(collect_weights)

    @staticmethod
    def compare(bench_dir, cmp_dir, output_path):
        bench_weight_pack_path = os.path.join(bench_dir, WeightsChecker.target_name_in_zip)
        cmp_weight_pack_path = os.path.join(cmp_dir, WeightsChecker.target_name_in_zip)
        df = compare_weight(bench_weight_pack_path, cmp_weight_pack_path)
        pass_check = Const.CONFIG_CHECK_PASS if False not in df['equal'].values else Const.CONFIG_CHECK_ERROR
        return WeightsChecker.target_name_in_zip, pass_check, df



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Union
from msprobe.infer.utils.check.checker import Checker, CheckResult, rule


class NumberChecker(Checker):

    @rule()
    def is_number(self) -> Union["NumberChecker", CheckResult]:
        return CheckResult(isinstance(self.instance, (int, float)), "not number")

    @rule()
    def is_int(self) -> Union["NumberChecker", CheckResult]:
        return CheckResult(isinstance(self.instance, int) and not isinstance(self.instance, bool), "not integer")

    @rule()
    def is_float(self) -> Union["NumberChecker", CheckResult]:
        return CheckResult(isinstance(self.instance, float), "not float")

    @rule()
    def is_zero(self) -> Union["NumberChecker", CheckResult]:
        return CheckResult(self.instance == 0, "not zero")

    @rule()
    def is_divisible_by(self, target) -> Union["NumberChecker", CheckResult]:
        if self.instance != 0:
            return CheckResult(target % self.instance == 0, f"not divisible by {target}")
        else:
            raise ValueError(f"Failed to check division, since {self.instance} is 0.")

    @rule()
    def in_range(self, min_value, max_value) -> Union["NumberChecker", CheckResult]:
        return self.greater_equal(min_value) and self.less_equal(max_value)

    @rule()
    def greater_equal(self, value) -> Union["NumberChecker", CheckResult]:
        return CheckResult(value <= self.instance, f"not greater equal to {value}")

    @rule()
    def less_equal(self, value) -> Union["NumberChecker", CheckResult]:
        return CheckResult(self.instance <= value, f"not less equal to {value}")

    @rule()
    def less_than(self, value) -> Union["NumberChecker", CheckResult]:
        return CheckResult(value > self.instance, f"not less than {value}")

    @rule()
    def greater_than(self, value) -> Union["NumberChecker", CheckResult]:
        return CheckResult(self.instance > value, f"not greater than {value}")



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This FusionOpComResult class. This file mainly involves the get_result function.
"""
import collections

from msprobe.msaccucmp.vector_cmp.fusion_manager import fusion_rule_parser
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.vector_cmp.range_manager.range_manager import RangeManager
from msprobe.msaccucmp.algorithm_manager.algorithm_manager import AlgorithmManager
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import FusionOp
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class TensorResult:
    """
    The class for tensor compare result
    """

    def __init__(self: any, tensor_info: dict, result: list, error_msg: list, is_ffts: bool) -> None:
        self.tensor_info = tensor_info
        self.algorithm_result = result[0]
        self.error_msg = error_msg
        self.overflow_result = result[1]
        self.is_ffts = is_ffts

    def get_result(self: any) -> list:
        """
        Get tensor result list
        :return [tensor_id, shape, algorithm_result, error_msg]
        """
        shape_str = '[%s]' % ",".join(map(str, self.tensor_info.get("shape", ConstManager.NAN)))
        if self.overflow_result:
            result = [shape_str] + [self.overflow_result] + self.algorithm_result + [",".join(self.error_msg)]
        else:
            result = [shape_str] + self.algorithm_result + [",".join(self.error_msg)]

        if self.tensor_info.get("tensor_id"):
            result = [self.tensor_info.get("tensor_id")] + result
        return result

    def get_algorithm_result(self: any) -> list:
        """
        Get algorithm result
        """
        return self.algorithm_result

    def get_my_output_dtype(self):
        return self.tensor_info.get("my_output_dtype", ConstManager.NAN)

    def get_ground_truth_dtype(self):
        return self.tensor_info.get("ground_truth_dtype", ConstManager.NAN)

    def get_my_output_address(self):
        return self.tensor_info.get("my_output_address", ConstManager.NAN)

    def get_ground_truth_address(self):
        return self.tensor_info.get("ground_truth_address", ConstManager.NAN)

    def get_op_type(self):
        return self.tensor_info.get("op_type", ConstManager.NAN)


class PytorchOpInfo:
    """
    The class for pytorch op info
    """

    def __init__(self: any, index: int, op_name: str, my_dump_path: str, ground_truth_dump_path: str) -> None:
        self.index = index
        self.op_name = op_name
        self.my_dump_path = my_dump_path
        self.ground_truth_dump_path = ground_truth_dump_path

    def get_result(self: any) -> list:
        """
        Get op info result
        :return [index, op_name, my_dump_path, ground_truth_dump_path]
        """
        return [str(self.index), self.op_name, self.my_dump_path, self.ground_truth_dump_path]

    def get_op_name(self: any) -> str:
        """
        Get op name
        """
        return self.op_name


class FusionOpComResult:
    """
    The class for fusion op compare result
    """

    def __init__(self: any, algorithm_manager: AlgorithmManager, ground_truth_to_my_output_map: any = None,
                 overflow_detection: bool = False, dump_is_cpu_or_gpu_data: list = None) -> None:
        self.algorithm_manager = algorithm_manager
        self.ground_truth_to_my_output_map = ground_truth_to_my_output_map
        self.overflow_detection = overflow_detection
        self.is_ground_truth_gpu_or_cpu = dump_is_cpu_or_gpu_data[0] if dump_is_cpu_or_gpu_data else False
        self.is_my_dump_gpu_or_cpu = dump_is_cpu_or_gpu_data[1] if dump_is_cpu_or_gpu_data else False

    @staticmethod
    def _make_ops_without_map(fusion_op: FusionOp, no_dump_file: bool) -> (str, str):
        my_output_op = fusion_op.op_name
        ground_truth_op = fusion_op.op_name
        # if only left or right has dump file
        if fusion_op.op_type in [ConstManager.LEFT_TYPE, ConstManager.RIGHT_TYPE]:
            if no_dump_file:
                if fusion_op.op_type == ConstManager.LEFT_TYPE:
                    ground_truth_op = '*'
                elif fusion_op.op_type == ConstManager.RIGHT_TYPE:
                    my_output_op = '*'
        else:
            ground_truth_op = ','.join(fusion_op.attr.original_op_names)
            if ground_truth_op == '':
                ground_truth_op = '*'
        return my_output_op, ground_truth_op

    @staticmethod
    def _process_input_and_output(result, input_result_list, output_result_list):
        if ConstManager.INPUT_PATTERN in result[ConstManager.TENSOR_INDEX]:
            input_result_list.append(result)
        elif ConstManager.OUTPUT_PATTERN in result[ConstManager.TENSOR_INDEX]:
            output_result_list.append(result)
        return input_result_list, output_result_list

    def get_result(self: any, fusion_op: FusionOp, tensor_result: any, error_msg: list,
                   no_dump_file: bool = False) -> any:
        """
        Get fusion op compare result list
        :param fusion_op: the fusion op
        :param tensor_result: the tensor result list
        :param error_msg: error message
        :param no_dump_file: the result no dump file
        :return [op_id, my_output_op, ground_truth_op, tensor_id, shape, algorithm_result, error_msg]
        """
        result_list = []
        input_result_list = []
        output_result_list = []
        is_ffts = False
        my_output_op, ground_truth_op = self._make_my_output_op_and_ground_truth_op(fusion_op, no_dump_file)
        if tensor_result:
            for item in tensor_result:
                current_tensor_info = [
                    str(fusion_op.op_id), item.get_op_type(),
                    my_output_op, str(item.get_my_output_dtype()),
                    str(item.get_my_output_address()), ground_truth_op,
                    str(item.get_ground_truth_dtype()), str(item.get_ground_truth_address())
                ]
                self._pre_handle_result(current_tensor_info)
                result = current_tensor_info + item.get_result()
                if item.is_ffts:
                    is_ffts = True
                    input_result_list, output_result_list = \
                        self._process_input_and_output(result, input_result_list, output_result_list)
                RangeManager.adjust_data(result, fusion_op.attr.get_op_sequence())
                log.print_info_log('[{}] Result: {}'.format(fusion_op.op_name, " ".join(result)))
                result_list.append(result)
        else:
            current_tensor_info = [
                str(fusion_op.op_id), fusion_op.get_real_op_type(),
                my_output_op, ConstManager.NAN, ConstManager.NAN,
                ground_truth_op, ConstManager.NAN, ConstManager.NAN,
                ConstManager.NAN, ConstManager.NAN
            ]
            self._pre_handle_result(current_tensor_info)
            if self.overflow_detection:
                # using 'NaN' as an overflow detection for 'no tensor_result'
                # and insert it after the column 'Shape'.
                result = current_tensor_info + ['NaN'] + self.algorithm_manager.make_nan_result() \
                         + [",".join(error_msg)]
            else:
                result = current_tensor_info + self.algorithm_manager.make_nan_result() + [",".join(error_msg)]
            RangeManager.adjust_data(result, fusion_op.attr.get_op_sequence())
            log.print_info_log('[{}] Result: {}'.format(fusion_op.op_name, " ".join(result)))
            result_list.append(result)

        Result = collections.namedtuple("Result", ["result_list", "input_result_list", "output_result_list", "is_ffts"])
        result = Result(result_list, input_result_list, output_result_list, is_ffts)
        return result

    def get_pytorch_result(self: any, op_info: PytorchOpInfo, tensor_result: any, error_msg: list) -> list:
        """
        Get fusion op compare result list
        :param op_info: the pytorch op info
        :param tensor_result: the tensor result list
        :param error_msg: error message
        :return [op_id, my_output_op, ground_truth_op, tensor_id, shape, algorithm_result, error_msg]
        """
        result_list = []
        if tensor_result:
            for item in tensor_result:
                result = op_info.get_result() + [item.get_my_output_dtype()] + item.get_result()
                log.print_info_log('[%s:%d] Result: %s' % (op_info.op_name, op_info.index, " ".join(result)))
                result_list.append(result)
        else:
            result = op_info.get_result() + [ConstManager.NAN] + [ConstManager.NAN] + \
                     self.algorithm_manager.make_nan_result() + [",".join(error_msg)]
            log.print_info_log('[%s:%d] Result: %s' % (op_info.op_name, op_info.index, " ".join(result)))
            result_list.append(result)
        return result_list

    def _make_my_output_op_and_ground_truth_op(self: any, fusion_op: FusionOp, no_dump_file: bool) -> (str, str):
        if self.ground_truth_to_my_output_map:
            my_output_op, ground_truth_op = fusion_rule_parser. \
                make_left_and_right_string(self.ground_truth_to_my_output_map)
        else:
            my_output_op, ground_truth_op = self._make_ops_without_map(fusion_op, no_dump_file)
        return my_output_op, ground_truth_op

    def _pre_handle_result(self: any, current_tensor_info: list) -> None:
        """
        if dump data is not NPU data, the result will be popped.
        the index of my dump data address is 3, the index of ground truth data address is 6.
        args: result list
        """
        if self.is_ground_truth_gpu_or_cpu:
            # op id is inserted as index in header, so address index should plus one.
            current_tensor_info.pop(ConstManager.GROUND_TRUTH_ADDRESS_INDEX + 1)
        if self.is_my_dump_gpu_or_cpu:
            current_tensor_info.pop(ConstManager.MY_OUTPUT_ADDRESS_INDEX + 1)


def get_result_title(algorithm_manager: AlgorithmManager, op_header: list, overflow_detection: bool = False) -> list:
    """
    Get result title
    :param algorithm_manager: the algorithm manager
    :param op_header: the op header
    :param overflow_detection: whether to display overflow info
    :return  [Index, op_header, shape, algorithm_result, error_msg]
    """

    header = ['Index'] + op_header + ['Shape'] + algorithm_manager.get_result_title() + ["CompareFailReason"]
    # add 'OverFlow' after 'Shape'.
    if overflow_detection:
        header.insert(header.index('Shape') + 1, 'OverFlow')
    RangeManager.adjust_header(header)
    return header


class SingleOpCmpResult:
    """
    The class for single op result
    """
    def __init__(self: any) -> None:
        self.op_name = ""
        self.dump_match = False
        self.result_list = None
        self.ret = 0
        self.input_list = None
        self.input_result_list = None
        self.output_result_list = None
        self.is_ffts = False
        self.op_name_origin_output_index_map = None
        self.npu_vs_npu = False

    @staticmethod
    def get_pre_op_output(op_name: str, index: int, result_mapping: dict) -> list:
        pre_op_result = result_mapping.get(op_name)
        if pre_op_result:
            output_result = pre_op_result.output_result_list[index]
        else:
            output_result = None
            message = "The result of '%s' is not in result mapping" % op_name
            log.print_warn_log(message)
        return output_result

    def update_attr(self: any, result_info: collections.namedtuple) -> None:
        self.op_name = result_info.op_name
        self.dump_match = result_info.dump_match
        self.result_list = result_info.result_list
        self.ret = result_info.ret
        self.input_list = result_info.input_list
        self.input_result_list = result_info.input_result_list
        self.output_result_list = result_info.output_result_list
        self.is_ffts = result_info.is_ffts
        self.op_name_origin_output_index_map = result_info.op_name_origin_output_index_map
        self.npu_vs_npu = result_info.npu_vs_npu

    def check_result_list_valid(self: any) -> None:
        if len(self.result_list) < len(self.input_result_list):
            message = "The length of input result list is greater than result list, '%s'" % self.op_name
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_INPUT_MAPPING)

    def find_pre_op(self: any, result_mapping: dict) -> None:
        """
        Replace the input of the current operator with the previous output result
        """
        self.check_result_list_valid()
        for index, input_result in enumerate(self.input_result_list):
            if len(input_result) <= ConstManager.TENSOR_INDEX:
                log.print_warn_log(f"Broken result, id {index}, skip")
                continue

            tensor_id = input_result[ConstManager.TENSOR_INDEX]
            pre_op = self.op_name_origin_output_index_map.get(tensor_id)
            if not pre_op or len(pre_op) < 2:
                message = "The tensor index '%s' is invalid, no input mapping information" % tensor_id
                log.print_error_log(message)
                raise CompareError(CompareError.MSACCUCMP_INVALID_INPUT_MAPPING)
            pre_op_name = pre_op[0]
            pre_op_index = pre_op[1]
            output_result = self.get_pre_op_output(pre_op_name, pre_op_index, result_mapping)
            if not output_result:
                continue
            origin_result = self.result_list[index][:ConstManager.TENSOR_INDEX + 1]
            self.result_list[index] = origin_result + output_result[ConstManager.TENSOR_INDEX + 1:]



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves convert model to json function.
"""
import os
import stat

from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import check_file_or_directory_path
from msprobe.infer.utils.util import filter_cmd
from msprobe.infer.offline.compare.msquickcmp.common.utils import execute_command, get_model_name_and_extension, \
    check_file_size_valid, ACCURACY_COMPARISON_MODEL_TYPE_ERROR, ACCURACY_COMPARISON_INVALID_PATH_ERROR, \
    MAX_READ_FILE_SIZE_4G, AccuracyCompareException

ATC_FILE_PATH = "compiler/bin/atc"
OLD_ATC_FILE_PATH = "atc/bin/atc"
NEW_ATC_FILE_PATH = "bin/atc"


def convert_model_to_json(cann_path, offline_model_path, out_path):
    """
    Function Description:
        convert om model to json
    Return Value:
        output json path
    Exception Description:
        when the model type is wrong throw exception
    """
    model_name, extension = get_model_name_and_extension(offline_model_path)
    if extension not in [".om", ".txt"]:
        logger.error(
            f"The offline model file not ends with .om or .txt, Please check {offline_model_path}.")
        raise AccuracyCompareException(ACCURACY_COMPARISON_MODEL_TYPE_ERROR)
    
    cann_path = os.path.realpath(cann_path)
    if not os.path.isdir(cann_path):
        logger.error(f'The cann path {cann_path} is not a directory.Please check.')
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
    
    atc_command_file_path = get_atc_path(cann_path)
    check_file_or_directory_path(atc_command_file_path)
    output_json_path = os.path.join(out_path, "model", model_name + ".json")
    if os.path.exists(output_json_path):
        logger.info(f"The {output_json_path} file is exists.")
    else:
        # do the atc command to convert om to json
        logger.info('Start to converting the model to json')
        if extension == ".om":
            mode_type = "1"
        else:
            mode_type = "5"
        atc_cmd = [
            atc_command_file_path, "--mode=" + mode_type, "--om=" + offline_model_path,
            "--json=" + output_json_path
        ]
        logger.info(f"ATC command line {' '.join(atc_cmd)}")
        atc_cmd = filter_cmd(atc_cmd)
        execute_command(atc_cmd)
        logger.info(f"Complete model conversion to json {output_json_path}.")

    check_file_size_valid(output_json_path, MAX_READ_FILE_SIZE_4G)
    return output_json_path


def get_atc_path(cann_path):
    atc_command_file_path = os.path.join(cann_path, ATC_FILE_PATH)
    if not os.path.exists(atc_command_file_path):
        atc_command_file_path = os.path.join(cann_path, OLD_ATC_FILE_PATH)
    if not os.path.exists(atc_command_file_path):
        atc_command_file_path = os.path.join(cann_path, NEW_ATC_FILE_PATH)
    if not os.path.exists(atc_command_file_path):
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)

    atc_command_file_path = os.path.realpath(atc_command_file_path)
    if not os.access(atc_command_file_path, os.X_OK):
        logger.error('ATC path is not permitted for executing.')
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
    if os.stat(atc_command_file_path).st_mode & (stat.S_IWGRP | stat.S_IWOTH) > 0:
        logger.error('ATC path is writable by others or group, not permitted.')
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
    return atc_command_file_path



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union
)
import os
import numpy as np
import mindspore
from mindspore import ops
from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import ApiAccuracyCheckerException
from msprobe.mindspore.api_accuracy_checker.compute_element import ComputeElement
from msprobe.mindspore.api_accuracy_checker.type_mapping import float_dtype_str_list, torch_dtype_to_dtype_str
from msprobe.mindspore.api_accuracy_checker.utils import convert_to_tuple
from msprobe.mindspore.api_accuracy_checker.bench_functions.fusion_operator import fusion
from msprobe.mindspore.common.const import MsCompareConst
from msprobe.mindspore.common.log import logger


from msprobe.mindspore.api_accuracy_checker import torch_mindtorch_importer

from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch
from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch_tensor
from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch_func
from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch_dist

if torch_mindtorch_importer.is_valid_pt_mt_env:
    from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import torch
else:
    import torch

# 为了可读性，我们先给每种返回形态起个别名
ForwardResult = Tuple[
    List[ComputeElement],
    Tuple[Any, ...],
    Dict[str, Any],
    Tuple[Any, ...],
]

BackwardResultMT = Tuple[
    List[ComputeElement],
    Union[Any, Tuple[Any, ...]],
    Tuple[Any, ...],
]

PyTorchBackward = List[ComputeElement]


class ApiInputAggregation:
    def __init__(self, inputs, kwargs, gradient_inputs) -> None:
        """
        Args:
            inputs: List[ComputeElement]
            kwargs: dict{str: ComputeElement}
            gradient_inputs: Union[List[ComputeElement], None]
        """
        self.inputs = inputs
        self.kwargs = kwargs
        self.gradient_inputs = gradient_inputs


api_parent_module_mapping = {
    (MsCompareConst.MINT, Const.MS_FRAMEWORK): mindspore.mint,
    (MsCompareConst.MINT, Const.PT_FRAMEWORK): torch,
    (MsCompareConst.MINT_FUNCTIONAL, Const.MS_FRAMEWORK): mindspore.mint.nn.functional,
    (MsCompareConst.MINT_FUNCTIONAL, Const.PT_FRAMEWORK): torch.nn.functional,
    (MsCompareConst.TENSOR_API, Const.MS_FRAMEWORK): mindspore.Tensor,
    (MsCompareConst.TENSOR_API, Const.PT_FRAMEWORK): torch.Tensor,
    (MsCompareConst.MINDTORCH_TENSOR, Const.MT_FRAMEWORK): mindtorch_tensor,
    (MsCompareConst.MINDTORCH_TENSOR, Const.PT_FRAMEWORK): torch.Tensor,
    (MsCompareConst.MINDTORCH, Const.MT_FRAMEWORK): mindtorch,
    (MsCompareConst.MINDTORCH, Const.PT_FRAMEWORK): torch,
    (MsCompareConst.MINDTORCH_FUNC, Const.MT_FRAMEWORK): mindtorch_func,
    (MsCompareConst.MINDTORCH_FUNC, Const.PT_FRAMEWORK): torch.nn.functional,
    (MsCompareConst.MINDTORCH_DIST, Const.MT_FRAMEWORK): mindtorch_dist,
    (MsCompareConst.MINDTORCH_DIST, Const.PT_FRAMEWORK): torch.distributed,
    (MsCompareConst.FUNCTIONAL_API, Const.MS_FRAMEWORK): mindspore.ops,
    (MsCompareConst.FUSION_API, Const.PT_FRAMEWORK): fusion

}


api_parent_module_str_mapping = {
    (MsCompareConst.MINT, Const.MS_FRAMEWORK): "mindspore.mint",
    (MsCompareConst.MINT, Const.PT_FRAMEWORK): "torch",
    (MsCompareConst.MINT_FUNCTIONAL, Const.MS_FRAMEWORK): "mindspore.mint.nn.functional",
    (MsCompareConst.MINT_FUNCTIONAL, Const.PT_FRAMEWORK): "torch.nn.functional",
    (MsCompareConst.TENSOR_API, Const.MS_FRAMEWORK): "mindspore.Tensor",
    (MsCompareConst.TENSOR_API, Const.PT_FRAMEWORK): "torch.Tensor",
    (MsCompareConst.MINDTORCH_TENSOR, Const.MT_FRAMEWORK): "mindtorch_tensor",
    (MsCompareConst.MINDTORCH_TENSOR, Const.PT_FRAMEWORK): "torch.Tensor",
    (MsCompareConst.MINDTORCH, Const.MT_FRAMEWORK): "mindtorch",
    (MsCompareConst.MINDTORCH, Const.PT_FRAMEWORK): "torch",
    (MsCompareConst.MINDTORCH_FUNC, Const.MT_FRAMEWORK): "mindtorch_func",
    (MsCompareConst.MINDTORCH_FUNC, Const.PT_FRAMEWORK): "torch.nn.functional",
    (MsCompareConst.MINDTORCH_DIST, Const.MT_FRAMEWORK): "mindtorch_dist",
    (MsCompareConst.MINDTORCH_DIST, Const.PT_FRAMEWORK): "torch.distributed",
    (MsCompareConst.FUNCTIONAL_API, Const.MS_FRAMEWORK): "mindspore.ops",
    (MsCompareConst.FUSION_API, Const.PT_FRAMEWORK): "fusion"
}


class ApiRunner:
    def __call__(self, api_input_aggregation, api_name_str, forward_or_backward=Const.FORWARD,
                 api_platform=Const.MS_FRAMEWORK):
        '''
        Args:
            api_input_aggregation: ApiInputAggregation
            api_name_str: str, e.g. "MintFunctional.relu.0"
            forward_or_backward: str, Union["forward", "backward"]
            api_platform: str, Union["mindspore", "torch", "mindtorch"]

        Return:
            outputs: list[ComputeElement]

        Description:
            run mindspore.mint/torch api
        '''

        api_type_str, api_sub_name = self.get_info_from_name(api_name_str, api_platform)
        api_instance = self.get_api_instance(api_type_str, api_sub_name, api_platform)

        return self.run_api(api_instance, api_input_aggregation, forward_or_backward, api_platform)

    @staticmethod
    def get_info_from_name(api_name_str, api_platform=Const.MS_FRAMEWORK):
        """
        Args:
            api_name_str: str, the trimmed key of data dict in api_info.json. e.g. "MintFunctional.relu.0"
            api_platform: str, the platform for the API, which can be either "mindspore" or "mindtorch".
                      It specifies which framework is being used. Default is "mindspore".
        Return:
            api_type_str: str, Union["MintFunctional", "Mint", "Tensor", "Torch", "Functional"]
            api_sub_name: str, e.g. "relu"
        """
        api_name_list = api_name_str.split(Const.SEP)
        if len(api_name_list) != 3:
            err_msg = f"ApiRunner.get_info_from_name failed: api_name_str: {api_name_str} is not in defined format." \
                      f" Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.WrongValue))
        api_type_str, api_sub_name = api_name_list[0], api_name_list[1]
        if api_type_str not in [MsCompareConst.MINT, MsCompareConst.MINT_FUNCTIONAL, MsCompareConst.TENSOR_API,
                                MsCompareConst.FUNCTIONAL_API] \
                and api_platform == Const.MS_FRAMEWORK:
            err_msg = f"ApiRunner.get_info_from_name failed: not mint, mint.nn.functional or Tensor api," \
                      f" api_name={api_name_str}. Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.WrongValue))

        if api_type_str not in MsCompareConst.MT_VALID_API_TYPES and api_platform == Const.MT_FRAMEWORK:
            err_msg = f"ApiRunner.get_info_from_name failed: not torch, functional or Tensor api," \
                      f" api_name={api_name_str}. Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.WrongValue))
        return api_type_str, api_sub_name

    @staticmethod
    def get_api_instance(api_type_str, api_sub_name, api_platform):
        """
        Args:
            api_type_str: str, Union["MintFunctional", "Mint", "Tensor", "Functional"]
            api_sub_name: str, e.g. "relu"
            api_platform: str: Union["mindspore", "pytorch"]

        Return:
            api_instance: function object

        Description:
            get mindspore.mint/torch api function
            mindspore.mint.{api_sub_name} <--> torch.{api_sub_name}
            mindspore.mint.nn.functional.{api_sub_name} <--> torch.nn.functional.{api_sub_name}
        """
        if api_sub_name in MsCompareConst.SUPPORTED_FUSION_LIST and api_platform == "pytorch":
            api_parent_module = api_parent_module_mapping.get((MsCompareConst.FUSION_API, api_platform))
            api_parent_module_str = api_parent_module_str_mapping.get((MsCompareConst.FUSION_API, api_platform))
        else:
            api_parent_module = api_parent_module_mapping.get((api_type_str, api_platform))
            api_parent_module_str = api_parent_module_str_mapping.get((api_type_str, api_platform))
        full_api_name = api_parent_module_str + Const.SEP + api_sub_name

        if not hasattr(api_parent_module, api_sub_name):
            err_msg = f"ApiRunner.get_api_instance failed: {full_api_name} is not found"
            logger.error_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.ApiWrong))

        api_instance = getattr(api_parent_module, api_sub_name)
        if not callable(api_instance):
            err_msg = f"ApiRunner.get_api_instance failed: {full_api_name} is not callable"
            logger.error_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.ApiWrong))

        return api_instance

    @staticmethod
    def run_api(
        api_instance,
        api_input_aggregation,
        forward_or_backward: str,
        api_platform: str,
    ) -> Union[ForwardResult, BackwardResultMT, PyTorchBackward]:
        inputs = tuple(compute_element.get_parameter(get_origin=False, tensor_platform=api_platform)
                       for compute_element in api_input_aggregation.inputs)
        kwargs = {key: value.get_parameter(get_origin=False, tensor_platform=api_platform)
                  for key, value in api_input_aggregation.kwargs.items()}
        gradient_inputs = api_input_aggregation.gradient_inputs

        if forward_or_backward == Const.FORWARD:
            forward_result = api_instance(*inputs, **kwargs)  # can be single tensor or tuple
            forward_result_tuple = convert_to_tuple(forward_result)
            res_compute_element_list = [ComputeElement(parameter=api_res) for api_res in forward_result_tuple]
            if api_platform == Const.MS_FRAMEWORK or api_platform == Const.MT_FRAMEWORK:
                return res_compute_element_list, inputs, kwargs, forward_result_tuple
        else:
            if gradient_inputs is None:
                err_msg = f"ApiRunner.run_api failed: run backward api but gradient_inputs is missing"
                logger.error_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.WrongValue))
            gradient_inputs = tuple(compute_element.get_parameter(get_origin=False, tensor_platform=api_platform)
                                    for compute_element in gradient_inputs)
            if api_platform == Const.MS_FRAMEWORK or api_platform == Const.MT_FRAMEWORK:
                if len(gradient_inputs) == 1:
                    gradient_inputs = gradient_inputs[0]

                def api_with_kwargs(*forward_inputs):
                    return api_instance(*forward_inputs, **kwargs)

                grad_func = ops.GradOperation(get_all=True, sens_param=True)(api_with_kwargs)
                backward_result = grad_func(*inputs, gradient_inputs)  # can be single tensor or tuple
                backward_result_tuple = convert_to_tuple(backward_result)
                res_compute_element_list = [ComputeElement(parameter=api_res) for api_res in backward_result_tuple]
                return res_compute_element_list, gradient_inputs, backward_result_tuple
            else:
                # set requires_grad
                requires_grad_index = []
                for index, tensor in enumerate(inputs):
                    if isinstance(tensor, torch.Tensor) and \
                            torch_dtype_to_dtype_str.get(tensor.dtype) in float_dtype_str_list:
                        setattr(tensor, "requires_grad", True)
                        requires_grad_index.append(index)
                forward_results = api_instance(*inputs, **kwargs)
                forward_results = convert_to_tuple(forward_results)
                for forward_res, gradient_in in zip(forward_results, gradient_inputs):
                    forward_res.backward(gradient_in)
                backward_result_list = []
                for index in requires_grad_index:
                    backward_result_list.append(getattr(inputs[index], "grad"))
                res_compute_element_list = [ComputeElement(parameter=api_res) for api_res in backward_result_list]

        return res_compute_element_list


api_runner = ApiRunner()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from msprobe.core.common.const import Const, MonitorConst
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.log import logger
from msprobe.core.monitor_v2.base import BaseMonitorV2
from msprobe.core.monitor_v2.utils import build_param2name, get_vpp_stage_from_tag, iter_model_chunks


class WeightGradMonitorV2(BaseMonitorV2, ABC):
    """
    Lightweight weight-gradient monitor that patches optimizer.step to grab
    pre/post gradients, reusing per-framework metric computation.
    """

    def __init__(self) -> None:
        super().__init__()
        self._optimizer: Any = None
        self._model: Any = None
        self._param2name: Dict[Any, str] = {}
        self._orig_step = None
        self._patched = False
        self._eps: float = 1e-8
        self._grad_hooks: List[Any] = []
        self.monitor_mbs_grad: bool = False
        self._micro_batch_number: int = 1
        self._param_micro_steps: Dict[Any, int] = {}

    @abstractmethod
    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        ...

    def set_config(self, config: Dict[str, Any]) -> None:
        super().set_config(config)
        self._eps = self._parse_eps(self.config.get("eps", 1e-8))
        self.monitor_mbs_grad = self._parse_bool(self.config.get("monitor_mbs_grad", False))

    def start(self, model: Any = None, optimizer: Any = None, **context: Any) -> None:
        self.set_context(**context)
        if self._optimizer is not None:
            return
        optimizer = optimizer or self._context.get("optimizer")
        if optimizer is None:
            raise ValueError("Optimizer must be provided to start weight grad function")
        self._optimizer = optimizer
        self._model = model or self._context.get("model")
        self._micro_batch_number = self._resolve_micro_batch_number()
        self._register_params(self._model, optimizer)
        self._patch_optimizer()

    def stop(self) -> None:
        if self._optimizer is None:
            return
        self._restore_optimizer()
        self._optimizer = None
        self._model = None
        self._rows = []
        self._param2name.clear()
        self._clear_grad_hooks()

    # ---------------------------------------------------------------------
    #  Hook registration helpers
    # ---------------------------------------------------------------------

    def _register_params(self, model: Any, optimizer: Any) -> None:
        self._param2name = build_param2name(model=model, optimizer=optimizer)
        self._register_pre_grad_hooks()

    def _register_pre_grad_hooks(self) -> None:
        self._param_micro_steps.clear()
        for param, name in self._param2name.items():
            if not getattr(param, "requires_grad", True):
                continue
            handle = self._create_param_hook(param, name)
            if handle is not None:
                self._grad_hooks.append(handle)

    def _create_param_hook(self, param: Any, name: str):
        hook_fn = getattr(param, "register_hook", None)
        if not callable(hook_fn):
            return None
        self._param_micro_steps[param] = 0
        return hook_fn(lambda grad, *, _p=param, _n=name: self._on_param_grad(_p, _n, grad))

    def _on_param_grad(self, param: Any, name: str, grad: Any) -> None:
        if grad is None or not FmkAdp.is_tensor(grad):
            return
        current_idx = self._param_micro_steps.get(param, 0) + 1
        self._param_micro_steps[param] = current_idx

        should_record = self.monitor_mbs_grad or (
            self._micro_batch_number <= 0 or current_idx >= self._micro_batch_number
        )
        if should_record:
            micro_step_val = current_idx if self.monitor_mbs_grad else self._micro_batch_number
            self._record_grad("unreduced", name, grad, micro_step=micro_step_val)

        if self._micro_batch_number > 0 and current_idx >= self._micro_batch_number:
            self._param_micro_steps[param] = 0

    def _clear_grad_hooks(self) -> None:
        for handle in self._grad_hooks:
            try:
                handle.remove()
            except Exception as exc:
                logger.warning(f"[monitor_v2] Failed to remove grad hook: {exc}")
        self._grad_hooks.clear()
        self._param_micro_steps.clear()

    # ---------------------------------------------------------------------
    #  Optimizer patching
    # ---------------------------------------------------------------------

    def _patch_optimizer(self) -> None:
        if self._patched or not hasattr(self._optimizer, "step"):
            return
        self._orig_step = self._optimizer.step

        def patched_step(*args: Any, **kwargs: Any):
            self._capture_grads("reduced")
            return self._orig_step(*args, **kwargs)

        self._optimizer.step = patched_step  # type: ignore[assignment]
        self._patched = True

    def _restore_optimizer(self) -> None:
        if self._patched and self._orig_step is not None:
            self._optimizer.step = self._orig_step  # type: ignore[assignment]
        self._patched = False
        self._orig_step = None
        self._clear_grad_hooks()

    # ---------------------------------------------------------------------
    #  Recording helpers
    # ---------------------------------------------------------------------

    def _capture_grads(self, kind: str) -> None:
        tag2tensor: Dict[str, Any] = {}
        tag2micro: Dict[str, int] = {}
        micro_step_val = self._micro_batch_number
        for param, name in self._param2name.items():
            grad = self._fetch_param_grad(param)
            if grad is None or not FmkAdp.is_tensor(grad):
                continue
            tag = name
            tag2tensor[tag] = grad
            tag2micro[tag] = micro_step_val
        self._record_grad_dict(kind, tag2tensor, tag2micro)

    def _resolve_micro_batch_number(self) -> int:
        candidates = [
            self.config.get("micro_batch_number"),
            self.config.get("grad_acc_steps"),
            self._context.get("micro_batch_number"),
            self._context.get("grad_acc_steps"),
        ]
        for candidate in candidates:
            if candidate is None:
                continue
            try:
                value = int(candidate)
            except (TypeError, ValueError):
                continue
            if value > 0:
                return value
        return 1

    def _record_grad(self, kind: str, name: str, grad: Any, micro_step: Optional[int] = None) -> None:
        if grad is None or not FmkAdp.is_tensor(grad):
            return
        tag = name
        micro_map = {tag: micro_step} if micro_step is not None else None
        self._record_grad_dict(kind, {tag: grad}, micro_map)

    def _record_grad_dict(
        self,
        kind: str,
        tag2tensor: Dict[str, Any],
        tag2micro: Optional[Dict[str, int]] = None,
    ) -> None:
        if not tag2tensor:
            return
        stats = self._compute_metrics(tag2tensor)
        for tag, op_dict in stats.items():
            vpp_stage = get_vpp_stage_from_tag(tag)
            row = {
                "module_name": tag,
                "scope": kind,
                "stats": {op: self._detach_stat(val) for op, val in op_dict.items()},
            }
            if tag2micro and tag in tag2micro:
                row["micro_step"] = tag2micro[tag]
            if vpp_stage is not None:
                row["vpp_stage"] = vpp_stage
            else:
                row["vpp_stage"] = MonitorConst.DEFAULT_STAGE
            self._rows.append(row)

    def _fetch_param_grad(self, param: Any) -> Any:
        return getattr(param, "grad", None)

    def _detach_stat(self, value: Any) -> Any:
        if value is None or not FmkAdp.is_tensor(value):
            return value
        if FmkAdp.fmk == Const.PT_FRAMEWORK:
            return value.detach()
        return value

    def _parse_eps(self, value: Any) -> float:
        try:
            eps = float(value)
        except (TypeError, ValueError):
            logger.warning(f"[monitor_v2] Invalid eps value: {value}, fallback to 1e-8.")
            return 1e-8
        if eps <= 0:
            logger.warning(f"[monitor_v2] eps must be positive, got {eps}, fallback to 1e-8.")
            return 1e-8
        return eps

    def _parse_bool(self, value: Any) -> bool:
        if isinstance(value, bool):
            return value
        logger.warning(f"[monitor_v2] Invalid boolean value: {value}, fallback to False.")
        return False




class PyTorchWeightGradMonitorV2(WeightGradMonitorV2):
    def __init__(self) -> None:
        super().__init__()
        self._fsdp_mode: bool = False
        self._orig_autograd_backward = None
        self._fsdp_backward_patched: bool = False
        self._fsdp_micro_step: int = 0

    def start(self, model: Any = None, optimizer: Any = None, **context: Any) -> None:
        self.set_context(**context)
        if self._optimizer is not None:
            return
        model = model or self._context.get("model")
        self._fsdp_mode = self._detect_fsdp(model)
        self._fsdp_micro_step = 0
        super().start(model=model, optimizer=optimizer, **context)
        if self._fsdp_mode:
            self._patch_autograd_backward()

    def stop(self) -> None:
        self._restore_autograd_backward()
        self._fsdp_mode = False
        self._fsdp_micro_step = 0
        super().stop()

    def _register_pre_grad_hooks(self) -> None:
        if self._fsdp_mode:
            self._param_micro_steps.clear()
            return
        super()._register_pre_grad_hooks()

    def _create_param_hook(self, param: Any, name: str):
        try:
            import torch

            param_tmp = param.expand_as(param)
            grad_acc = param_tmp.grad_fn.next_functions[0][0]
            if grad_acc is None:
                raise RuntimeError
        except Exception:
            return super()._create_param_hook(param, name)

        @torch.no_grad()
        def acc_hook(*_unused):
            grad = self._fetch_param_grad(param)
            self._on_param_grad(param, name, grad)

        self._param_micro_steps[param] = 0
        return grad_acc.register_hook(acc_hook)

    def _fetch_param_grad(self, param: Any) -> Any:
        grad = getattr(param, "main_grad", None)
        if grad is None:
            grad = getattr(param, "grad", None)
        if grad is None:
            return None
        if grad.__class__.__name__ == "DTensor":
            try:
                grad = grad.to_local()
            except Exception as exc:
                logger.warning(f"[monitor_v2] Failed to convert DTensor grad to local: {exc}")
        try:
            from msprobe.pytorch.common.utils import is_float8_tensor

            if is_float8_tensor(grad):
                grad = grad.float()
        except Exception as exc:
            logger.warning(f"[monitor_v2] Failed to convert grad tensor from float8: {exc}")
        try:
            return grad.clone()
        except Exception:
            return grad

    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        from msprobe.pytorch.monitor.module_metric import get_metrics

        return get_metrics(self._ops, tag2tensor, self._eps, {})

    def _detect_fsdp(self, model: Any) -> bool:
        if model is None:
            return False
        try:
            from torch.distributed.fsdp import FullyShardedDataParallel as FSDP  # type: ignore[import]
        except Exception:
            FSDP = None
        for chunk, _prefix in iter_model_chunks(model):
            if FSDP is not None and isinstance(chunk, FSDP):
                if getattr(chunk, "_use_orig_params", False):
                    return True
            try:
                for _param_name, param in chunk.named_parameters():
                    if param.__class__.__name__ == "DTensor":
                        return True
            except Exception:
                continue
        return False

    def _patch_autograd_backward(self) -> None:
        if self._fsdp_backward_patched:
            return
        try:
            import torch.autograd as _autograd
        except Exception:
            logger.warning("[monitor_v2] Failed to import torch.autograd, skipping FSDP grad patch.")
            return

        orig = getattr(_autograd, "backward", None)
        if orig is None:
            return

        def wrapped_backward(*args: Any, **kwargs: Any):
            out = orig(*args, **kwargs)
            self._capture_fsdp_pre_grads()
            return out

        self._orig_autograd_backward = orig
        _autograd.backward = wrapped_backward
        self._fsdp_backward_patched = True

    def _restore_autograd_backward(self) -> None:
        if not self._fsdp_backward_patched or self._orig_autograd_backward is None:
            self._fsdp_backward_patched = False
            self._orig_autograd_backward = None
            return
        try:
            import torch.autograd as _autograd

            _autograd.backward = self._orig_autograd_backward
        except Exception as exc:
            logger.warning(f"[monitor_v2] Failed to restore torch.autograd.backward: {exc}")
        self._fsdp_backward_patched = False
        self._orig_autograd_backward = None

    def _capture_fsdp_pre_grads(self) -> None:
        if self._model is None:
            return
        current_idx = self._fsdp_micro_step + 1
        self._fsdp_micro_step = current_idx

        tag2tensor: Dict[str, Any] = {}
        tag2micro: Dict[str, int] = {}
        should_record = True
        if self.monitor_mbs_grad:
            micro_step_val = current_idx
        else:
            micro_step_val = self._micro_batch_number
            if self._micro_batch_number > 0 and current_idx < self._micro_batch_number:
                should_record = False

        if should_record:
            for param, name in self._param2name.items():
                if not getattr(param, "requires_grad", True):
                    continue
                grad = self._fetch_param_grad(param)
                if grad is None or not FmkAdp.is_tensor(grad):
                    continue
                clean_name = name.replace("_fsdp_wrapped_module.", "")
                tag = clean_name
                tag2tensor[tag] = grad
                tag2micro[tag] = micro_step_val

            self._record_grad_dict("unreduced", tag2tensor, tag2micro)

        if self._micro_batch_number > 0 and current_idx >= self._micro_batch_number:
            self._fsdp_micro_step = 0


class MindSporeWeightGradMonitorV2(WeightGradMonitorV2):
    def _compute_metrics(self, tag2tensor: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        from msprobe.mindspore.monitor.utils import get_metrics

        return get_metrics(self._ops, tag2tensor, self._eps, {})



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the overflow analyse function.
"""
import argparse
import os
import time

import numpy as np

from msprobe.msaccucmp.cmp_utils import log, utils_type, path_check
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.dump_parse.dump_data_parser import DumpDataParser
from msprobe.msaccucmp.cmp_utils.file_utils import OverflowFileUtils, DumpFileDesc


class OverflowAnalyse:
    """
    The class for parse the overflow info
    """

    OVER_FLOW_TYPE = [
        'AI Core',
        'DHA Atomic Add',
        'L2 Atomic Add'
    ]

    def __init__(self: any, args: any = None) -> None:
        self.dump_path = os.path.realpath(args.dump_path)
        if os.path.islink(os.path.abspath(args.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(args.output_path)
        self.top_n = args.top_num
        self.debug_files = None
        self.overflow_file_utils = OverflowFileUtils()

    @staticmethod
    def check_argument(arguments: any = None) -> int:
        """
        check the arguments of overflow
        :arguments: input args
        """
        ret = path_check.check_path_valid(arguments.dump_path, exist=True,
                                     path_type=path_check.PathType.Directory)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            log.print_error_log('[Overflow] the -d parameter: "%r"'
                                ' is invalid!' % arguments.dump_path)
        else:
            ret = path_check.check_output_path_valid(arguments.output_path, exist=True)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                log.print_error_log('[Overflow] the -out parameter: "%r"'
                                    ' is invalid!' % arguments.output_path)
        return ret

    @staticmethod
    def npy_data_summary(source_data: any) -> str:
        """
        Get npy information
        :source_data: np data
        """
        if isinstance(source_data, str):
            if not str(source_data).endswith('.npy'):
                raise CompareError(
                    CompareError.MSACCUCMP_INVALID_TYPE_ERROR)
            data = np.load(source_data)
        elif isinstance(source_data, np.ndarray):
            data = source_data
        else:
            raise CompareError(
                CompareError.MSACCUCMP_INVALID_TYPE_ERROR)
        if np.size(data) == 0:
            raise CompareError(
                CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)
        return '[Shape: %s] [Dtype: %s] [Max: %s] [Min: %s] [Mean: %s]' \
               % (data.shape, data.dtype, np.max(data), np.min(data), data.mean())

    @staticmethod
    def _get_overflow_info_new_version(res: list, json_txt: any) -> any:
        acc_type = json_txt.get('acc_list', {}).get('acc_type')
        if acc_type in ConstManager.ACC_TYPE.values():
            overflow_type = acc_type
            detail = json_txt.get('acc_list', {}).get('data', {})
            return OverflowAnalyse._gen_overflow_info(res, overflow_type, detail)
        log.print_error_log("[Overflow] Invalid overflow type, type is {}".format(acc_type))
        raise CompareError(CompareError.MSACCUCMP_INVALID_OVERFLOW_TYPE_ERROR)

    @staticmethod
    def _get_overflow_info_old_version(res: list, json_txt: any) -> any:
        overflow = False
        id_info = ()
        for overflow_type, detail in json_txt.items():
            if detail.get('status'):
                overflow = True
                id_info = OverflowAnalyse._gen_overflow_info(res, overflow_type, detail)

        if not overflow:
            log.print_error_log("[Overflow] Invalid json_txt, overflow type invalid or status is zero!")
            raise CompareError(CompareError.MSACCUCMP_INVALID_OVERFLOW_TYPE_ERROR)
        return id_info

    @staticmethod
    def _gen_overflow_info(res: list, overflow_type: str, detail: any) -> any:
        status = detail.get('status')
        if status and status != 0:
            overflow_info = ' [%s][TaskId:%s][StreamId:%s][Status:%s]' \
                            % (overflow_type, detail.get('task_id'),
                               detail.get('stream_id'), status)
            res.append(overflow_info)
            task_info = (
                detail.get('task_id'),
                detail.get('stream_id'),
                detail.setdefault('context_id', ConstManager.INVALID_ID),
                detail.setdefault('thread_id', ConstManager.INVALID_ID)
            )
            return task_info
        log.print_error_log("[Overflow] The OpDebug file exists, but the value of status is {}!".format(status))
        raise CompareError(CompareError.MSACCUCMP_INVALID_OVERFLOW_STATUS_ERROR)

    @staticmethod
    def _insert_delimiter(res: list, over_index: int) -> list:
        """
        insert delimiter for every overflow result
        :over_index: the overflow index
        """
        res.insert(0, '=================================================[%d]'
                      '==================================================' % over_index)
        return res

    @staticmethod
    def _parse_overflow_file(file_path: str, output_path: str) -> any:
        """
        parse debug file or dump/debug file
        :file_path: the path of dump/debug file
        :output_path: store the result file
        """
        args = argparse.Namespace(dump_path=file_path,
                                  output_path=output_path,
                                  dump_version=2,
                                  output_file_type='npy')
        return DumpDataParser(args).parse_dump_data()

    @staticmethod
    def _is_dump_file_match(dump_file_desc: DumpFileDesc, task_info: any):
        task_id = task_info[0]
        stream_id = task_info[1]
        context_id = task_info[2] if task_info[2] != ConstManager.INVALID_ID else None
        thread_id = task_info[3] if task_info[3] != ConstManager.INVALID_ID else None
        if dump_file_desc.task_id != task_id or dump_file_desc.stream_id != stream_id:
            return False
        if context_id and dump_file_desc.context_id != context_id:
            return False
        if thread_id and dump_file_desc.thread_id != thread_id:
            return False
        return True

    def analyse(self: any) -> int:
        """
        analyse overflow info
        """
        if self._find_all_debug_files():
            max_num = len(self.debug_files)
            if self.top_n < len(self.debug_files):
                max_num = self.top_n
            log.print_info_log('[Overflow] Find [{}] overflow ops. Will show the top {}.'
                               .format(len(self.debug_files), max_num))

            overflow_result = ''
            for i, debug_file in enumerate(self.debug_files):
                if i >= max_num:
                    break
                parsed_debug_file = self._get_parsed_debug_file(debug_file)
                overflow_json = self.overflow_file_utils.load_json_file(parsed_debug_file.file_path)
                if not isinstance(overflow_json, dict):
                    log.print_warn_log("[Overflow] overflow summary file {} contents is not dict"
                                       .format(parsed_debug_file.file_path))
                    continue
                overflow_result = '{}{}\n'.format(overflow_result,
                                                  self._json_summary(i + 1, overflow_json, debug_file))

            log.print_info_log("[Overflow] The overflow analyse result:\n %s" % overflow_result)
            result_file_name = 'overflow_summary_%s.txt' \
                               % time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
            self.overflow_file_utils.save_file(os.path.join(self.output_path, result_file_name),
                                               overflow_result)

        return CompareError.MSACCUCMP_NONE_ERROR

    def _find_all_debug_files(self: any) -> bool:
        """
        find debug_files and sort it
        """
        debug_files = self.overflow_file_utils.list_dump_files(self.dump_path,
                                                               self.overflow_file_utils.DEBUG_FILE_PATTERN)
        # sort by timestamp
        self.debug_files = sorted(debug_files, key=lambda x: x.timestamp)
        if len(self.debug_files) == 0:
            log.print_warn_log("[Overflow] Find [0] overflow node!")
            return False
        return True

    def _json_summary(self: any, overflow_index: int, json_txt: any, debug_file: any) -> any:
        """
        get the summary info about overflow op
        :overflow_index: index of overflow op
        :json_txt: the parsed info from debug file
        :debug_file: the desc of debug file
        """
        res = []
        if ConstManager.MAGIC_KEY_WORD in json_txt.keys():
            task_info = self._get_overflow_info_new_version(res, json_txt)
        else:
            task_info = self._get_overflow_info_old_version(res, json_txt)
        res.append(' [timestamp:%s]' % debug_file.timestamp)
        try:
            dump_file_desc = self._find_dump_files_by_task_id(os.path.dirname(debug_file.file_path),
                                                              task_info)
        except CompareError:
            log.print_warn_log("[Overflow] Can't find the dump file corresponding to the"
                               " debug file: %s" % os.path.basename(debug_file.file_path))
            res = self._insert_delimiter(res, overflow_index)
            return '\n'.join(res)
        finally:
            pass

        parsed_dump_files = self._get_parsed_dump_file(dump_file_desc)
        # sort input/output & index
        for anchor_type in ['input', 'output']:
            for parsed_dump_file in parsed_dump_files:
                if parsed_dump_file.type != anchor_type:
                    continue
                res.append(' %s' % os.path.basename(parsed_dump_file.file_path))
                res.append(' -[Format: %s] %s' % (parsed_dump_file.format,
                                                  self.npy_data_summary(parsed_dump_file.file_path)))

        res.insert(0, '[%s] %s' % (dump_file_desc.op_type, dump_file_desc.op_name))
        res = self._insert_delimiter(res, overflow_index)
        return "\n".join(res)

    def _get_parsed_debug_file(self: any, debug_file_desc: any) -> any:
        """
        get parsed debug file
        :debug_file_desc: the DumpFileDesc of the debug file
        """
        file_name = os.path.basename(debug_file_desc.file_path)
        parsed_debug_files = self.overflow_file_utils.list_parsed_debug_files(self.output_path, file_name)
        if not parsed_debug_files:
            self._parse_overflow_file(debug_file_desc.file_path, self.output_path)
            parsed_debug_files = self.overflow_file_utils.list_parsed_debug_files(self.output_path, file_name)
            if not parsed_debug_files:
                log.print_warn_log("[Overflow] Parsed overflow debug file: %s failed." % file_name)
                raise CompareError(CompareError.MSACCUCMP_PARSE_DUMP_FILE_ERROR)
        # only one json file when parse one debug file
        if len(parsed_debug_files) != 1:
            raise CompareError(CompareError.MSACCUCMP_MATCH_MORE_FILE_ERROR)
        return list(parsed_debug_files.values())[0]

    def _get_parsed_dump_file(self: any, dump_file_desc: any) -> list:
        """
        get parsed dump file
        :dump_file_desc: the DumpFileDesc of the dump file
        """
        self._parse_overflow_file(dump_file_desc.file_path, self.output_path)
        parsed_dump_files = self.overflow_file_utils.list_parsed_dump_files(self.output_path, dump_file_desc)
        if len(parsed_dump_files) == 0:
            log.print_warn_log("[Overflow] Parsed overflow dump file: %s failed."
                               % os.path.basename(dump_file_desc.file_path))
            raise CompareError(CompareError.MSACCUCMP_UNKNOWN_ERROR)
        return sorted(parsed_dump_files.values(), key=lambda x: x.idx)

    def _find_dump_files_by_task_id(self: any, dump_path: str, task_info: any) -> any:
        """
        find dump file by the task id and stream id
        :dump_path：the dump file path
        :task_id: the task id for the dump file
        :stream_id: the steam id for the dump file
        """
        dump_files = self.overflow_file_utils.list_dump_files(dump_path, self.overflow_file_utils.DUMP_FILE_PATTERN)
        dump_file_list = []
        for item in dump_files:
            if item.op_type != 'Opdebug' \
                    and self._is_dump_file_match(item, task_info):
                dump_file_list.append(item)
        if dump_file_list:
            dump_file_list.sort(key=lambda x: x.timestamp)
            return dump_file_list[0]
        raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR)



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the dump function.
"""
import os
import time
import struct
import warnings
import ctypes
import tempfile
import json
from typing.io import BinaryIO

import numpy as np
from google.protobuf.message import DecodeError

from msprobe.msaccucmp.dump_parse.proto_dump_data import DumpData
from msprobe.msaccucmp.cmp_utils import path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class BigDumpDataParser:
    """
    The class for big dump data parser
    """
    warnings.filterwarnings("ignore")

    def __init__(self: any, dump_file_path: str) -> None:
        self.dump_file_path = dump_file_path
        self.file_size = 0
        self.header_length = 0
        self.dump_data = DumpData()
        self.dump_json_data = {}
        self.data_types = ['input', 'output', 'buffer', 'space']
        self.parse_dump_so = "libascend_dump_parser.so"

    @staticmethod
    def find_shared_library(lib_name: str, relative_path: str) -> str:
        """
        优先从LD_LIBRARY_PATH查找 so 文件，如果没有则从相对路径中查找。
        """
        # 从环境变量 LD_LIBRARY_PATH 中查找
        ld_paths = os.environ.get("LD_LIBRARY_PATH", "").split(":")
        for path in ld_paths:
            candidate = os.path.join(path, lib_name)
            if os.path.exists(candidate):
                return os.path.realpath(candidate)

        current_dir = os.path.dirname(os.path.realpath(__file__))
        candidate = os.path.realpath(os.path.join(current_dir, relative_path, lib_name))

        if os.path.exists(candidate):
            return candidate

        raise OSError(
            f"Shared library {lib_name} not found."
        )

    def parse(self: any):
        """
        Parse the dump file path by big dump data format
        :return: DumpData
        :exception when read or parse file error
        """
        self.check_argument_valid()
        try:
            with open(self.dump_file_path, 'rb') as dump_file:
                # read header length
                self._read_header_length(dump_file)
                self._parse_dump_to_json()
                self._parse_binary_to_json_data(dump_file)
                return self.dump_data.from_dict(self.dump_json_data)
        except (OSError, IOError) as io_error:
            log.print_error_log('Failed to read the dump file %r. %s'
                                % (self.dump_file_path, str(io_error)))
            raise CompareError(CompareError.MSACCUCMP_OPEN_FILE_ERROR) from io_error
        finally:
            pass

    def check_argument_valid(self: any) -> None:
        """
        check argument valid
        :exception when invalid
        """
        ret = path_check.check_path_valid(self.dump_file_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        # get file size
        try:
            self.file_size = os.path.getsize(self.dump_file_path)
        except (OSError, IOError) as error:
            log.print_error_log('get the size of dump file %r failed. %s'
                                % (self.dump_file_path, str(error)))
            raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR) from error
        finally:
            pass

        if self.file_size <= ConstManager.UINT64_SIZE:
            log.print_error_log(
                'The size of %r must be greater than %d, but the file size'
                ' is %d. Please check the dump file.'
                % (self.dump_file_path, ConstManager.UINT64_SIZE, self.file_size))
            raise CompareError(CompareError.MSACCUCMP_UNMATCH_STANDARD_DUMP_SIZE)
        if self.file_size > ConstManager.ONE_GB:
            log.print_warn_log(
                'The size (%d) of %r exceeds 1GB, it may task more time to run, please wait.'
                % (self.file_size, self.dump_file_path))

    def _read_header_length(self: any, dump_file: BinaryIO) -> None:
        # read header length
        header_length = dump_file.read(ConstManager.UINT64_SIZE)
        self.header_length = struct.unpack(ConstManager.UINT64_FMT, header_length)[0]
        # check header_length <= file_size - 8
        if self.header_length > self.file_size - ConstManager.UINT64_SIZE:
            log.print_warn_log(
                'The header content size (%d) of %r must be less than or'
                ' equal to %d (file size) - %d (header length).'
                ' Please check the dump file.'
                % (self.header_length, self.dump_file_path, self.file_size, ConstManager.UINT64_SIZE))
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)
        dump_file.read(self.header_length)

    def _parse_dump_to_json(self):
        # 读取 dump 二进制数据
        with open(self.dump_file_path, 'rb') as dump_file:
            binary_data = dump_file.read()
        # 加载 C 解析库
        try:
            self.parse_dump_so = self.find_shared_library(self.parse_dump_so, "../../../adump/lib64")
            ret = path_check.check_path_valid(self.parse_dump_so, True, False)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)
            if path_check.is_group_and_others_writable(self.parse_dump_so):
                log.print_error_log(f"Failed to load {self.parse_dump_so}, this file is not safe. Please check.")
                raise CompareError(CompareError.MSACCUCMP_INVALID_FILE_ERROR)
            dump_parse_cdll = ctypes.CDLL(self.parse_dump_so)
        except (OSError, IOError) as e:
            log.print_error_log(f"Failed to load {self.parse_dump_so}:{e}")
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR) from e
        data_ptr = ctypes.c_char_p(binary_data)

        # 使用临时文件替代落盘 JSON
        fd, tmp_json_path = tempfile.mkstemp(suffix=".json")
        # 关闭 fd，让 C 库写文件
        os.close(fd)
        try:
            res = dump_parse_cdll.ParseDumpProtoToJson(
                data_ptr, ctypes.c_size_t(len(binary_data)), tmp_json_path.encode('utf-8'))

            if res != 0 or not os.path.isfile(tmp_json_path):
                log.print_error_log(f"Parse dump file to json failed.")
                raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)
            # 从临时文件加载 JSON
            with open(tmp_json_path, 'r') as load_f:
                self.dump_json_data = json.load(load_f)
        finally:
            # 删除临时文件，避免落盘
            if os.path.exists(tmp_json_path):
                os.remove(tmp_json_path)

    def _parse_binary_to_json_data(self, dump_file: BinaryIO):
        used_size = self.header_length + ConstManager.UINT64_SIZE
        for data_type in self.data_types:
            for item in self.dump_json_data.get(data_type, []):
                size = int(item.get('size', 0))
                used_size += size
                if used_size > self.file_size:
                    log.print_error_log(f'The size of {self.dump_file_path} is invalid, please check the dump file.')
                    raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)
                item['data'] = dump_file.read(size)


class DumpDataHandler:
    """
    Handle dump data
    """

    def __init__(self: any, dump_file_path: str) -> None:
        self.dump_file_path = dump_file_path
        self.file_size = 0

    def check_argument_valid(self: any) -> None:
        """
        check argument valid
        :exception when invalid
        """
        ret = path_check.check_path_valid(self.dump_file_path, True)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        # get file size
        try:
            self.file_size = os.path.getsize(self.dump_file_path)
        except (OSError, IOError) as error:
            log.print_error_log('get the size of dump file %r failed. %s'
                                % (self.dump_file_path, str(error)))
            raise CompareError(CompareError.MSACCUCMP_DUMP_FILE_ERROR) from error
        finally:
            pass
        if self.file_size == 0:
            message = 'Failed to parse dump file %r. The file size is zero. Please check the dump file.' \
                      % self.dump_file_path
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message)
        if self.file_size > ConstManager.ONE_GB:
            log.print_warn_log(
                'The size (%d) of %r exceeds 1GB, it may task more time to run, please wait.'
                % (self.file_size, self.dump_file_path))

    def read_numpy_file(self: any) -> np.ndarray:
        """
        Read numpy file
        :return: numpy data
        """
        self.check_argument_valid()
        try:
            if self.dump_file_path.endswith(".txt"):
                numpy_data = np.loadtxt(self.dump_file_path)
            else:
                numpy_data = np.load(self.dump_file_path)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError, DecodeError) as error:
            log.print_error_log('Failed to parse dump file "%r". Only data of the numpy format is supported. %s'
                                % (self.dump_file_path, str(error)))
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR) from error
        finally:
            pass
        return numpy_data

    def parse_dump_data(self: any):
        """
        Parse dump file
        :param dump_version: the dump version
        :return: DumpData
        """
        self.check_argument_valid()
        if self.dump_file_path.endswith('.npy'):
            numpy_data = self.read_numpy_file()
            dump_data, _ = _convert_numpy_to_dump(numpy_data, only_header=False)
            return dump_data
        try:
            with open(self.dump_file_path, 'rb') as dump_file:
                file_content = dump_file.read()
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            message = 'Failed to open dump file %r. Please check the dump file. %s' \
                      % (self.dump_file_path, str(error))
            log.print_error_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message) from error
        finally:
            pass
        return BigDumpDataParser(self.dump_file_path).parse()


def _convert_numpy_to_dump(numpy_data: np.ndarray, only_header: bool):
    dump_data = DumpData()
    dump_data.version = '2.0'
    dump_data.dump_time = int(round(time.time() * ConstManager.TIME_LENGTH))
    output = dump_data.output.add()
    output.data_type = common.get_data_type_by_dtype(numpy_data.dtype)
    output.format = DD.FORMAT_RESERVED
    total_size = 1
    for dim in numpy_data.shape:
        output.shape.dim.append(dim)
        total_size *= dim
    # make output data
    struct_format = common.get_struct_format_by_data_type(output.data_type)
    data = struct.pack('%d%s' % (total_size, struct_format), *numpy_data.flatten())
    output.size = len(output.data)
    if not only_header:
        output.data = data
    return dump_data, data


def write_dump_data(numpy_data: np.ndarray, output_dump_path: str) -> None:
    """
    write numpy data to dump data file
    :param numpy_data: the numpy data
    :param output_dump_path: the output dump file path
    :exception when write file error
    """
    # make content of dump data header
    dump_data, data = _convert_numpy_to_dump(numpy_data, only_header=True)
    dump_data_ser = dump_data.SerializeToString()
    try:
        path_check.check_write_path_secure(output_dump_path)
        with os.fdopen(os.open(output_dump_path, ConstManager.WRITE_FLAGS,
                               ConstManager.WRITE_MODES), 'wb') as dump_file:
            # write the header length
            dump_file.write(struct.pack(ConstManager.UINT64_FMT, len(dump_data_ser)))
            # write the header content
            dump_file.write(dump_data_ser)
            # write output data
            dump_file.write(data)
    except IOError as io_error:
        log.print_error_log('Failed to write dump file %r. %s'
                            % (output_dump_path, str(io_error)))
        raise CompareError(CompareError.MSACCUCMP_WRITE_FILE_ERROR) from io_error
    finally:
        pass



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from functools import wraps
from typing import Any, Callable

import torch
from torch.utils.hooks import BackwardHook

from msprobe.core.common.const import Const
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.utils import is_float8_tensor
from msprobe.pytorch.dump.api_dump.api_register import get_api_register

torch_version_above_or_equal_2 = torch.__version__.split('+')[0] >= '2.0'


def wrap_setup_backward_hook(func):
    def requires_clone(tensor, need_check_leaf=False):
        if not isinstance(tensor, torch.Tensor):
            return False
        need_clone = torch.is_grad_enabled() and tensor.requires_grad and not is_float8_tensor(tensor)
        if need_check_leaf:
            need_clone &= tensor.grad_fn is not None
        return need_clone

    @recursion_depth_decorator("Dump: wrap_setup_backward_hook.parse_tensor", max_depth=Const.DUMP_MAX_DEPTH)
    def parse_tensor(item, tensor_list):
        if requires_clone(item):
            tensor_list.append(item)
        elif isinstance(item, (list, tuple)):
            for value in item:
                parse_tensor(value, tensor_list)
        elif isinstance(item, dict):
            for value in item.values():
                parse_tensor(value, tensor_list)

    @recursion_depth_decorator("Dump: wrap_setup_backward_hook.rebuild_args", max_depth=Const.DUMP_MAX_DEPTH)
    def rebuild_args(item, tensor_iter, need_check_leaf=False):
        if requires_clone(item, need_check_leaf):
            result = next(tensor_iter)
            if hasattr(result, "_base") and result._base is not None:
                if torch._C._autograd._get_creation_meta(result) != torch._C._autograd.CreationMeta(0):
                    torch._C._autograd._set_creation_meta(result, torch._C._autograd.CreationMeta(0))
            return result
        if isinstance(item, list):
            for index, value in enumerate(item):
                item[index] = rebuild_args(value, tensor_iter, need_check_leaf=True)
            return item
        if isinstance(item, dict):
            for key, value in item.items():
                item[key] = rebuild_args(value, tensor_iter, need_check_leaf=True)
            return item
        if isinstance(item, tuple):
            if hasattr(item, '_fields'):
                return type(item)(*[rebuild_args(i, tensor_iter) for i in item])
            return type(item)([rebuild_args(i, tensor_iter) for i in item])
        return item

    @wraps(func)
    def wrap_setup_hook_func(*args, **kwargs):
        if len(args) < 2:
            return func(*args, **kwargs)

        actual_args = args[1]

        tensor_list = []

        parse_tensor(actual_args, tensor_list)

        new_args = args[0], tuple(tensor_list)
        hooked_tensors = func(*new_args, **kwargs)

        tensor_iter = iter(hooked_tensors)
        try:
            new_data = rebuild_args(actual_args, tensor_iter)
        except Exception as e:
            logger.debug(f"Unsupported data in setup input/output hook. The detail info: {e}")
            new_data = actual_args

        return new_data

    return wrap_setup_hook_func


def wrap_setup_input_output_hook():
    BackwardHook.setup_input_hook = wrap_setup_backward_hook(BackwardHook.setup_input_hook)
    BackwardHook.setup_output_hook = wrap_setup_backward_hook(BackwardHook.setup_output_hook)


def get_apply_func_wrapper(original_func: Callable) -> Callable:
    @wraps(original_func)
    def wrapped_apply(*args, **kwargs) -> Any:
        api_register = get_api_register()
        if api_register:
            api_register.restore_inner_used_api()
        result = original_func(*args, **kwargs)
        if api_register:
            api_register.register_inner_used_api()
        return result

    return wrapped_apply


def wrap_backward_hook_function_apply():
    if torch_version_above_or_equal_2:
        original_apply = torch.nn.modules._functions.BackwardHookFunction.apply
        torch.nn.modules._functions.BackwardHookFunction.apply = get_apply_func_wrapper(original_apply)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

import mindspore
import numpy as np
from mindspore._c_expression import typing
from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import ApiAccuracyCheckerException
from msprobe.core.common.file_utils import load_npy
from msprobe.mindspore.api_accuracy_checker.type_mapping import (api_info_type_str_to_type,
                                                                 ms_dtype_to_dtype_str, torch_dtype_to_dtype_str,
                                                                 dtype_str_to_ms_dtype, dtype_str_to_np_dtype,
                                                                 dtype_str_to_mindtorch_dtype,
                                                                 dtype_str_to_torch_dtype, type_to_api_info_type_str,
                                                                 DEFAULT_CONSTRUCT_NP_FLOAT_DTYPE, TUPLE_TYPE_STR,
                                                                 MINDSPORE_TENSOR_TYPE_STR, MINDSPORE_DTYPE_TYPE_STR,
                                                                 SLICE_TYPE_STR, TORCH_DTYPE_TYPE_STR,
                                                                 float_dtype_str_list, int_dtype_str_list)
from msprobe.mindspore.api_accuracy_checker.utils import check_and_get_from_json_dict, global_context
from msprobe.mindspore.common.log import logger

import msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer as env_module


if env_module.is_valid_pt_mt_env:
    from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch
    from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import torch
else:
    import torch


class MstensorMetaData:
    def __init__(self, dtype_str, npy_path, maximum, minimum, shape) -> None:
        self.dtype_str = dtype_str
        self.npy_path = npy_path
        self.maximum = maximum
        self.minimum = minimum
        self.shape = shape


class DtypeMetaData:
    def __init__(self, dtype_str) -> None:
        self.dtype_str = dtype_str


class ComputeElement:
    def __init__(self, compute_element_info=None, parameter=None):
        self.supported_parameter_type = tuple(type_to_api_info_type_str.keys()) + tuple([torch.Tensor, tuple])
        if parameter is not None:
            self._init_with_parameter(parameter)
        elif isinstance(compute_element_info, (list, dict)):
            self._init_from_compute_element_info(compute_element_info)
        elif compute_element_info is None:
            self._init_from_null_compute_element_info()
        else:
            logger.warning_log_with_exp(
                "ComputeElement.__init__ failed: not init with parameter or compute_element info is not (list, dict)."
                " Exception has been raised and will be captured/logged externally.",
                ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))

    @staticmethod
    def transfer_to_torch_tensor(ms_tensor):
        '''
        Args:
            ms_tensor: mindspore.Tensor
        Return:
            torch_tensor: torch.Tensor
        '''
        ms_dtype = ms_tensor.dtype
        dtype_str = ms_dtype_to_dtype_str.get(ms_dtype)
        if dtype_str not in dtype_str_to_torch_dtype:
            err_msg = f"ComputeElement.transfer_to_torch_tensor failed: no matching torch dtype" \
                      f" for {dtype_str}. Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))
        else:
            torch_dtype = dtype_str_to_torch_dtype.get(dtype_str)

        if dtype_str in int_dtype_str_list:
            middle_dtype = mindspore.int64
        else:
            middle_dtype = mindspore.float64
        np_ndarray = ms_tensor.astype(middle_dtype).numpy()
        torch_tensor = torch.from_numpy(np_ndarray).to(torch_dtype)
        return torch_tensor

    @staticmethod
    def transfer_to_mindtorch_tensor(ms_tensor):
        """
        Args:
            ms_tensor: mindspore.Tensor
        Return:
            mindtorch_tensor: mindtorch.Tensor
        """

        ms_dtype = ms_tensor.dtype

        dtype_str = ms_dtype_to_dtype_str.get(ms_dtype)

        if dtype_str not in dtype_str_to_mindtorch_dtype:
            err_msg = f"ComputeElement.transfer_to_mindtorch_tensor failed: no matching mindtorch dtype for" \
                      f" {dtype_str}. Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg,
                                      ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))
        else:
            mindtorch_dtype = dtype_str_to_mindtorch_dtype.get(dtype_str)

        if dtype_str in int_dtype_str_list:
            middle_dtype = mindspore.int64
        else:
            middle_dtype = mindspore.float64

        np_ndarray = ms_tensor.astype(middle_dtype).numpy()

        mindtorch_tensor = mindtorch.from_numpy(np_ndarray).to(ms_dtype)

        return mindtorch_tensor

    @staticmethod
    def transfer_to_mindspore_tensor(torch_tensor):
        '''
        Args:
            torch_tensor: torch.Tensor

        Return:
            ms_tensor: mindspore.Tensor
        '''
        torch_dtype = torch_tensor.dtype
        dtype_str = torch_dtype_to_dtype_str.get(torch_dtype)
        if dtype_str not in dtype_str_to_ms_dtype:
            err_msg = \
                f"ComputeElement._transfer_to_mindspore_tensor failed: no matching mindspore dtype for {dtype_str}. " \
                f"Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))
        else:
            ms_dtype = dtype_str_to_ms_dtype.get(dtype_str)

        if dtype_str in int_dtype_str_list:
            middle_dtype = torch.int64
        else:
            middle_dtype = torch.float64
        np_ndarray = torch_tensor.to(middle_dtype, copy=True).numpy()
        ms_tensor = mindspore.Tensor.from_numpy(np_ndarray).astype(ms_dtype)
        return ms_tensor

    @staticmethod
    def convert_inf_to_real_num(value, dtype_str):
        if value == float("inf"):
            np_dtype = dtype_str_to_np_dtype.get(dtype_str, DEFAULT_CONSTRUCT_NP_FLOAT_DTYPE)
            value = np.finfo(np_dtype).max
        elif value == float("-inf"):
            np_dtype = dtype_str_to_np_dtype.get(dtype_str, DEFAULT_CONSTRUCT_NP_FLOAT_DTYPE)
            value = np.finfo(np_dtype).min
        return value

    def get_parameter(self, get_origin=True, tensor_platform=Const.MS_FRAMEWORK):
        '''
        Args:
            get_origin: boolean
            tensor_platform: str, Union["mindspore", "pytorch"]

        Return:
            parameter: Union[int, float, str, slice, tuple, torch.Tensor, mindspore.Tensor]
        '''
        if self.parameter is None:
            return self.parameter
        if isinstance(self.parameter, tuple):
            return tuple([compute_element.get_parameter(get_origin=get_origin, tensor_platform=tensor_platform)
                          for compute_element in self.parameter])
        elif isinstance(self.parameter, self.supported_parameter_type):
            parameter_tmp = self.parameter
        elif isinstance(self.parameter, DtypeMetaData):
            if tensor_platform == Const.MS_FRAMEWORK:
                parameter_tmp = dtype_str_to_ms_dtype.get(self.parameter.dtype_str)
            elif tensor_platform == Const.PT_FRAMEWORK:
                parameter_tmp = dtype_str_to_torch_dtype.get(self.parameter.dtype_str)
            elif tensor_platform == Const.MT_FRAMEWORK:
                parameter_tmp = dtype_str_to_mindtorch_dtype.get(self.parameter.dtype_str)

        elif isinstance(self.parameter, MstensorMetaData):
            mstensor_meta_data = self.parameter
            ms_dtype = dtype_str_to_ms_dtype.get(mstensor_meta_data.dtype_str)
            if global_context.get_is_constructed():
                np_dtype = dtype_str_to_np_dtype.get(mstensor_meta_data.dtype_str, DEFAULT_CONSTRUCT_NP_FLOAT_DTYPE)
                ndarray = self._construct_ndarray(mstensor_meta_data.shape, mstensor_meta_data.maximum,
                                                  mstensor_meta_data.minimum, np_dtype)
            else:
                ndarray = load_npy(mstensor_meta_data.npy_path)
            parameter_tmp = mindspore.Tensor(ndarray, dtype=ms_dtype)
        else:
            err_msg = "ComputeElement.get_parameter failed: self.parameter type is not in " \
                      "(int, float, str, slice, bool, torch.Tensor, mindspore.Tensor, MstensorMetaData)." \
                      "Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))

        # if necessary, do transfer
        if not get_origin and isinstance(parameter_tmp, mindspore.Tensor) and tensor_platform == Const.PT_FRAMEWORK:
            parameter = self.transfer_to_torch_tensor(parameter_tmp)
        elif not get_origin and isinstance(parameter_tmp, mindspore.Tensor) and tensor_platform == Const.MT_FRAMEWORK:
            parameter = self.transfer_to_mindtorch_tensor(parameter_tmp)
        elif not get_origin and isinstance(parameter_tmp, torch.Tensor) and tensor_platform == Const.MS_FRAMEWORK:
            parameter = self.transfer_to_mindspore_tensor(parameter_tmp)
        else:
            parameter = parameter_tmp

        return parameter

    def get_shape(self):
        return self.shape

    def get_dtype(self):
        return self.dtype_str

    def _construct_ndarray(self, shape, maximum, minimum, np_dtype):
        shape = tuple(shape)
        np.random.seed(42)
        if np_dtype == np.bool_:
            ndarray = np.random.rand(*shape) > 0.5
        else:
            maximum = self.convert_inf_to_real_num(maximum, np_dtype)
            minimum = self.convert_inf_to_real_num(minimum, np_dtype)
            ndarray = np.random.uniform(minimum, maximum, shape).astype(np_dtype)
        return ndarray

    def _init_from_null_compute_element_info(self):
        self.parameter = None
        self.shape = tuple()
        self.dtype = "None"

    def _init_from_compute_element_info(self, compute_element_info):
        '''
        Args:
            compute_element_info: Union[list, dict]

        Return:
            void

        init member attributes: self.shape, self.dtype_str, self.parameter
        '''
        if isinstance(compute_element_info, list):
            self.shape = tuple()
            self.dtype_str = TUPLE_TYPE_STR
            self.parameter = tuple([ComputeElement(compute_element_info=sub_info)
                                    for sub_info in compute_element_info])
        else:
            type_str = check_and_get_from_json_dict(compute_element_info, "type", "type field in api_info.json",
                                                    accepted_type=str, accepted_value=api_info_type_str_to_type.keys())
            self.shape = tuple()
            self.dtype_str = type_str
            if type_str == MINDSPORE_TENSOR_TYPE_STR:
                self._init_from_mstensor_compute_element_info(compute_element_info)
            else:
                value = check_and_get_from_json_dict(compute_element_info, "value", "value field in api_info.json")
                if type_str == MINDSPORE_DTYPE_TYPE_STR:
                    self.parameter = DtypeMetaData(value)
                elif type_str == SLICE_TYPE_STR:
                    self.parameter = slice(*tuple(value))
                else:  # type_str in ("str", "int", "float", "bool")
                    self.parameter = value

    def _init_from_mstensor_compute_element_info(self, compute_element_info):
        '''
            do not load real tensor, only record meta data
        '''
        dtype_str = check_and_get_from_json_dict(compute_element_info, "dtype", "dtype field in api_info.json",
                                                 accepted_type=str, accepted_value=dtype_str_to_ms_dtype.keys())
        shape = check_and_get_from_json_dict(compute_element_info, "shape", "shape field in api_info.json",
                                             accepted_type=(list,))
        if global_context.get_is_constructed():
            maximum = check_and_get_from_json_dict(compute_element_info, "Max", "Max field in api_info.json",
                                                   accepted_type=(int, float))
            minimum = check_and_get_from_json_dict(compute_element_info, "Min", "Min field in api_info.json",
                                                   accepted_type=(int, float))

            npy_path = None
        else:
            maximum, minimum = None, None
            data_name = check_and_get_from_json_dict(compute_element_info, "data_name",
                                                     "data_name field in api_info.json", accepted_type=(str,))
            npy_path = os.path.join(global_context.get_dump_data_dir(), data_name)
        mstensor_meta_data = MstensorMetaData(dtype_str, npy_path, maximum, minimum, shape)
        self.parameter = mstensor_meta_data
        self.dtype_str = dtype_str
        self.shape = tuple(shape)

    def _init_with_parameter(self, parameter):
        self.parameter = parameter
        self.shape = tuple()
        if not isinstance(parameter, self.supported_parameter_type):
            err_msg = "ComputeElement._init_with_parameter failed: " \
                      "parameter type is not in (int, float, str, slice, bool, torch.Tensor, mindspore.Tensor)." \
                      "Exception has been raised and will be captured/logged externally."
            logger.warning_log_with_exp(err_msg, ApiAccuracyCheckerException(ApiAccuracyCheckerException.UnsupportType))
        if isinstance(parameter, mindspore.Tensor):
            self.shape = tuple(parameter.shape)
            self.dtype_str = ms_dtype_to_dtype_str.get(parameter.dtype)
        elif isinstance(parameter, torch.Tensor):
            self.shape = tuple(parameter.shape)
            self.dtype_str = torch_dtype_to_dtype_str.get(parameter.dtype)
        elif isinstance(parameter, typing.Type):
            self.dtype_str = MINDSPORE_DTYPE_TYPE_STR
            self.parameter = DtypeMetaData(ms_dtype_to_dtype_str.get(parameter))
        elif isinstance(parameter, torch.dtype):
            self.dtype_str = TORCH_DTYPE_TYPE_STR
            self.parameter = DtypeMetaData(torch_dtype_to_dtype_str.get(parameter))
        elif isinstance(parameter, tuple):
            self.dtype_str = TUPLE_TYPE_STR
            self.parameter = tuple([ComputeElement(parameter=param) for param in parameter])
        else:
            self.dtype_str = type_to_api_info_type_str.get(type(parameter))



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
AlgorithmManager class.
This class mainly involves the compare function.
"""
import os
import re
import sys
import time
import stat
import importlib

import numpy as np

from msprobe.msaccucmp.cmp_utils import log, utils_type
from msprobe.msaccucmp.cmp_utils import utils, path_check
from msprobe.msaccucmp.algorithm_manager.algorithm_parameter import AlgorithmParameter
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils
from msprobe.msaccucmp.dump_parse import dump_utils


class AlgorithmManager:
    """
    The class for algorithm manager
    """

    def __init__(self: any, custom_script_path: str, select_algorithm: object, algorithm_options: str) -> None:
        self.custom_path = custom_script_path
        self.built_in_support_algorithm = []
        self.custom_support_algorithm = []
        self._make_support_algorithm()
        self.select_algorithm_list = self._parse_selection_algorithm(select_algorithm)
        self.algorithm_param = self._parse_algorithm_argument(algorithm_options)
        self.support_algorithm_map = self._make_select_algorithm_map()

    @staticmethod
    def _check_value_invalid(value: object, parameter: str) -> None:
        is_list_value_invalid = isinstance(value, list) and len(value) != 2
        if not value or is_list_value_invalid:
            log.print_error_log('The algorithm argument (%r) is invalid, just supports '
                                '"algorithm_name:name1=value1,name2=value2;".' % parameter)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    @staticmethod
    def _update_algorithm_param(algorithm_name: str, algorithm_param: dict, param_map: dict) -> None:
        if RegManager.match_pattern(RegManager.BUILTIN_ALGORITHM_INDEX_PATTERN, algorithm_name):
            algorithm_name = ConstManager.BUILT_IN_ALGORITHM[int(algorithm_name)]
        if algorithm_name in algorithm_param:
            algorithm_param[algorithm_name] = dict(algorithm_param.get(algorithm_name), **param_map)
        else:
            algorithm_param[algorithm_name] = param_map

    @staticmethod
    def _get_function(algorithm_module: any, module_type: str) -> (bool, any):
        algorithm_func = getattr(algorithm_module, ConstManager.COMPARE_FUNC_NAME)
        # check compare is function
        if not callable(algorithm_func):
            log.print_warn_log("[%s] The '%s' in %s is not function. Please check the file." %
                               (module_type, ConstManager.COMPARE_FUNC_NAME, str(algorithm_module.__file__)))
            return False, ''

        # check argument count of compare
        if algorithm_func.__code__.co_argcount != ConstManager.COMPARE_ARGUMENT_COUNT:
            log.print_warn_log("[%s] The argument count (%d) of '%s' in %s is not %d. Please check the file." %
                               (module_type, algorithm_func.__code__.co_argcount, ConstManager.COMPARE_FUNC_NAME,
                                str(algorithm_module.__file__), ConstManager.COMPARE_ARGUMENT_COUNT))
            return False, ''
        return True, algorithm_func

    @staticmethod
    def _add_algorithm_file_to_list(file_path: str, support_algorithm_list: list) -> bool:
        if os.path.isfile(file_path):
            file_name_pattern = re.compile(ConstManager.ALGORITHM_FILE_NAME_PATTERN)
            match = file_name_pattern.match(os.path.basename(file_path))
            if match is not None:
                file_stat = os.stat(file_path)
                file_mode = file_stat.st_mode
                # 判断others或group权限是否有写权限
                if bool(file_mode & stat.S_IWGRP):
                    log.print_warn_log(f"File {file_path} is not safe. Groups have writing permission to this file.")
                if bool(file_mode & stat.S_IWOTH):
                    log.print_error_log(f"File {file_path} is dangerous. Others have writing "
                                        "permission to this file. Please use chmod to dismiss the writing permission.")
                    raise CompareError(CompareError.MSACCUCMP_DANGER_FILE_ERROR)
                support_algorithm_list.append(match.group(1))
                current_uid = os.getuid()
                # 判断当前用户是普通用户情况下，文件创建者是否是当前用户
                if current_uid != 0 and file_stat.st_uid != 0 and file_stat.st_uid != current_uid:
                    log.print_error_log(f"File {file_path} is not owned by current user, "
                                        "if must use this file, copy or chmod this file by yourself.")
                    raise CompareError(CompareError.MSACCUCMP_DANGER_FILE_ERROR)
                return True
            log.print_warn_log("The file '%r' does not match 'alg_{algorithm_name}.py'"
                               " in '%r', please check the file." % (os.path.basename(file_path),
                                                                     os.path.dirname(file_path)))
        return False

    @staticmethod
    def _check_data_size_valid(my_output_dump_data: any, ground_truth_dump_data: any, args: dict) -> None:
        my_output_dump_data_size = len(my_output_dump_data)
        if my_output_dump_data_size == 0 and args.get('my_output_dump_file') is not None:
            msg = 'The dump data size is 0 in %r.' % args.get('my_output_dump_file')
            log.print_warn_log(msg)
        ground_truth_dump_data_size = len(ground_truth_dump_data)
        if ground_truth_dump_data_size == 0 and args.get('ground_truth_dump_file') is not None:
            msg = 'The dump data size is 0 in %r.' % args.get('ground_truth_dump_file')
            log.print_warn_log(msg)

        if my_output_dump_data_size != ground_truth_dump_data_size \
                and args.get('my_output_dump_file') is not None \
                and args.get('ground_truth_dump_file') is not None:
            msg = "The my output dump data size (%d) in '%r' does not match the ground truth dump data size (%d) " \
                  "in '%r'." % (my_output_dump_data_size, args.get('my_output_dump_file'),
                                ground_truth_dump_data_size, args.get('ground_truth_dump_file'))
            log.print_warn_log(msg)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, msg)

    @staticmethod
    def _check_return_value_valid(value: object, select_algorithm: str) -> None:
        # check the return value is string
        if not isinstance(value, str):
            err_msg = "The return value (%s) of '%s' in '%s' is not string. Please check the return value." \
                      % (value, ConstManager.COMPARE_FUNC_NAME, select_algorithm)
            raise CompareError(ConstManager.COMPARE_FUNC_NAME, err_msg)

    def compare(self: any, my_output_dump_data: any, ground_truth_dump_data: any,
                args: dict, max_cmp_size: int = 0) -> (list, list):
        """
        Compare my output dump data and the ground truth dump data by select algorithm
        :param my_output_dump_data: my output dump data to compare
        :param ground_truth_dump_data: the ground truth dump data to compare
        :param args: the algorithm parameter
        :param max_cmp_size: max cmp array size
        :return the result list and compare_fail_message
        """
        self._check_data_size_valid(my_output_dump_data, ground_truth_dump_data, args)
        np.seterr(divide='ignore', invalid='ignore')
        result = []
        error_msg = []

        is_bool_data = my_output_dump_data.dtype == np.bool_ and ground_truth_dump_data.dtype == np.bool_
        if max_cmp_size:
            my_output_dump_data_to_cmp = my_output_dump_data[:max_cmp_size]
            ground_truth_dump_data_to_cmp = ground_truth_dump_data[:max_cmp_size]
        else:
            my_output_dump_data_to_cmp = my_output_dump_data
            ground_truth_dump_data_to_cmp = ground_truth_dump_data

        if not is_bool_data:
            my_output_dump_data_to_cmp = np.array(my_output_dump_data_to_cmp).astype("float32")
            ground_truth_dump_data_to_cmp = np.array(ground_truth_dump_data_to_cmp).astype("float32")

        for select_algorithm, compare_func in self.support_algorithm_map.items():
            if is_bool_data and select_algorithm not in ConstManager.BOOL_ALGORITHM:
                result.append(ConstManager.NAN)
                error_msg += ["Algorithm %s does not support Boolean types." % select_algorithm]
                continue

            alg_args = self._make_algorithm_param(select_algorithm, args)

            # call compare function
            alg_result, alg_error_msg = self._call_compare_function(
                compare_func, my_output_dump_data_to_cmp, ground_truth_dump_data_to_cmp, alg_args, select_algorithm)
            result.append(alg_result)
            if alg_error_msg:
                error_msg += [alg_error_msg]
        return result, error_msg

    def get_result_title(self: any) -> list:
        """
        Get algorithm name list
        :return: the list
        """
        return list(self.support_algorithm_map.keys())

    def make_nan_result(self: any) -> list:
        """
        Make nan result for compare algorithm
        :return: the list, result is nan
        """
        return [ConstManager.NAN] * len(self.support_algorithm_map)

    def _add_algorithm_name_to_list(self: any, algorithm_name: str, select_algorithm_list: list) -> None:
        if not algorithm_name:
            return
        if RegManager.match_pattern(RegManager.BUILTIN_ALGORITHM_INDEX_PATTERN, algorithm_name):
            if ConstManager.BUILT_IN_ALGORITHM[int(algorithm_name)] not in select_algorithm_list:
                select_algorithm_list.append(ConstManager.BUILT_IN_ALGORITHM[int(algorithm_name)])
            return
        if algorithm_name not in self.built_in_support_algorithm + self.custom_support_algorithm:
            log.print_warn_log("The '%s' does not supported in builtin or custom. Please check the "
                               "algorithm name or algorithm index." % algorithm_name)
            return
        if algorithm_name not in select_algorithm_list:
            select_algorithm_list.append(algorithm_name)

    def _make_all_algorithm_list(self: any, select_algorithm: str) -> list:
        select_algorithm_list = []
        if select_algorithm.lower() == 'all':
            select_algorithm_list = ConstManager.BUILT_IN_ALGORITHM
            for item in self.custom_support_algorithm:
                if item not in self.custom_support_algorithm:
                    select_algorithm_list.append(item)
            return select_algorithm_list
        return select_algorithm_list

    def _parse_selection_algorithm(self: any, select_algorithm: object) -> list:
        if not select_algorithm:
            log.print_error_log('There is no algorithm to select. Please select at least one algorithm.')
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        if isinstance(select_algorithm, list):
            select_algorithm = ','.join(str(x) for x in select_algorithm)

        select_algorithm_list = self._make_all_algorithm_list(select_algorithm)
        if select_algorithm_list:
            return select_algorithm_list
        for item in select_algorithm.split(','):
            self._add_algorithm_name_to_list(item.strip(), select_algorithm_list)
        if not select_algorithm_list:
            log.print_error_log(
                "The algorithm in '%s' does not supported. Just supports %s. Please check the "
                "select algorithm name." % (select_algorithm,
                                            self.built_in_support_algorithm + self.custom_support_algorithm))
            raise CompareError(CompareError.MSACCUCMP_INVALID_ALGORITHM_ERROR)
        return select_algorithm_list

    def _change_str_to_map(self: any, algorithm_param_info: list, algorithm_args: str) -> dict:
        param_map = {}
        for param_str in algorithm_param_info:
            if param_str.strip():
                param = param_str.split('=')
                self._check_value_invalid(param, algorithm_args)
                self._check_value_invalid(param[0].strip(), algorithm_args)
                self._check_value_invalid(param[1].strip(), algorithm_args)
                param_map[param[0].strip()] = param[1].strip()
        return param_map

    def _parse_algorithm_argument(self: any, algorithm_options: str) -> dict:
        algorithm_param = {}
        if not algorithm_options:
            return algorithm_param

        for algorithm_args in algorithm_options.split(';'):
            if algorithm_args:
                algorithm_param_info = algorithm_args.split(':')
                self._check_value_invalid(algorithm_param_info, algorithm_args)
                algorithm_name = algorithm_param_info[0].strip()
                self._check_value_invalid(algorithm_name, algorithm_args)
                param_map = self._change_str_to_map(algorithm_param_info[1].split(','), algorithm_args)
                self._update_algorithm_param(algorithm_name, algorithm_param, param_map)
        return algorithm_param

    def _make_support_algorithm(self: any) -> None:
        """
        Make support algorithm list for custom and built in
        """
        dir_path = os.path.join(os.path.dirname(__file__), ConstManager.BUILT_IN_ALGORITHM_DIR_NAME)
        log.print_info_log("dir_path:%s" % dir_path)
        self._make_support_algorithm_by_path(dir_path, self.built_in_support_algorithm)
        if self.custom_path:
            ret = path_check.check_path_valid(
                self.custom_path, True, False, path_check.PathType.Directory)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)
            dir_path = os.path.join(self.custom_path, ConstManager.CUSTOM_ALGORITHM_DIR_NAME)
            path_check.check_path_all_file_exec_valid(dir_path)
            self._make_support_algorithm_by_path(dir_path, self.custom_support_algorithm)

    def _get_module(self: any, algorithm_name: str) -> (bool, any, str):
        if algorithm_name in self.custom_support_algorithm:
            sys.path.append(self.custom_path)
            algorithm_module = importlib.import_module('%s.alg_%s' %
                                                       (ConstManager.CUSTOM_ALGORITHM_DIR_NAME,
                                                        algorithm_name))
            module_type = ConstManager.CUSTOM
        elif algorithm_name in self.built_in_support_algorithm:
            algorithm_module = importlib.import_module('%s.%s.alg_%s' %
                                                       ("msprobe.msaccucmp.algorithm_manager",
                                                        ConstManager.BUILT_IN_ALGORITHM_DIR_NAME,
                                                        algorithm_name))
            module_type = ConstManager.BUILTIN
        else:
            return False, '', ''
        # check exist compare attr
        if not hasattr(algorithm_module, ConstManager.COMPARE_FUNC_NAME):
            log.print_warn_log("[%s] The file '%s' has no attribute '%s'. Please check the file."
                               % (module_type, str(algorithm_module.__file__), ConstManager.COMPARE_FUNC_NAME))
            return False, algorithm_module, module_type
        return True, algorithm_module, module_type

    def _make_select_algorithm_map(self: any) -> dict:
        """
        Make support algorithm map by select algorithm list
        """
        support_algorithm_map = {}
        for select_algorithm in self.select_algorithm_list:
            get_ok, algorithm_module, module_type = self._get_module(select_algorithm)
            if not get_ok:
                continue
            get_ok, algorithm_func = self._get_function(algorithm_module, module_type)
            if not get_ok:
                continue
            support_algorithm_map[select_algorithm] = algorithm_func
        if not support_algorithm_map:
            log.print_error_log(
                "The algorithm in %s does not supported. Just supports %s. Please check the "
                "select algorithm name." % (self.select_algorithm_list,
                                            self.built_in_support_algorithm + self.custom_support_algorithm))
            raise CompareError(CompareError.MSACCUCMP_INVALID_ALGORITHM_ERROR)
        return support_algorithm_map

    def _make_support_algorithm_by_path(self: any, dir_path: str, support_algorithm_list: list) -> None:
        if not os.path.exists(dir_path):
            log.print_warn_log("There is no '%s' in '%r', please check the custom path."
                               % (ConstManager.CUSTOM_ALGORITHM_DIR_NAME, os.path.dirname(dir_path)))
            return
        one_match = False
        for item in os.listdir(dir_path):
            if self._add_algorithm_file_to_list(os.path.join(dir_path, item), support_algorithm_list):
                one_match = True
        if not one_match:
            log.print_warn_log("There is no legal 'alg_{algorithm_name}.py' file in '%r', "
                               "please check the path." % dir_path)

    def _make_algorithm_param(self: any, algorithm_name: str, args: dict) -> AlgorithmParameter:
        alg_arg_map = args
        if algorithm_name in self.algorithm_param:
            alg_arg_map = dict(self.algorithm_param.get(algorithm_name), **args)
        alg_args = AlgorithmParameter()
        alg_args.__dict__ = alg_arg_map
        return alg_args

    def _call_compare_function(self: any, *args: any) -> (str, str):
        compare_func, my_output_dump_data, ground_truth_dump_data, alg_args, algorithm_name = args
        alg_result = ''
        alg_error_msg = ''
        try:
            if compare_func:
                alg_result, alg_error_msg = compare_func(my_output_dump_data, ground_truth_dump_data, alg_args)
                self._check_return_value_valid(alg_result, algorithm_name)
                self._check_return_value_valid(alg_error_msg, algorithm_name)
        except Exception as ex:
            alg_error_msg = "Failed to execute '%s' in '%s'. %s" \
                            % (ConstManager.COMPARE_FUNC_NAME, algorithm_name, str(ex))
            log.print_warn_log(alg_error_msg)
            alg_result = ConstManager.NAN
        finally:
            pass
        return alg_result, alg_error_msg


class AlgorithmManagerMain:
    """
    The class for algorithm manager main
    """

    def __init__(self: any, args: any) -> None:
        self.my_output_dump_file_path = os.path.realpath(args.my_dump_path)
        self.ground_truth_dump_file_path = os.path.realpath(args.golden_dump_path)
        self.manager = AlgorithmManager(args.custom_script_path, args.algorithm, args.algorithm_options)
        if args.output_path:
            if os.path.islink(os.path.abspath(args.output_path)):
                log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_path)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
            self.output_path = os.path.realpath(args.output_path)

    def check_arguments_valid(self: any) -> None:
        """
        Check arguments valid, if invalid, throw exception
        """
        ret = path_check.check_path_valid(self.my_output_dump_file_path, True, False,
                                          path_check.PathType.File)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        ret = path_check.check_path_valid(self.ground_truth_dump_file_path, True, False,
                                          path_check.PathType.File)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

    def process(self: any, save_result: bool = False) -> int:
        """
        Do compare for two numpy file
        :param save_result: is save result
        :return: VectorComparisonErrorCode
        """
        self.check_arguments_valid()
        log.print_info_log("The my output dump file is %r." % self.my_output_dump_file_path)
        log.print_info_log("The ground truth file is %r." % self.ground_truth_dump_file_path)
        try:
            self._process_exec(save_result)
        except CompareError as error:
            return error.code
        finally:
            pass
        return CompareError.MSACCUCMP_NONE_ERROR

    def _process_exec(self: any, save_result: bool) -> None:
        my_output_dump_data = dump_utils.read_numpy_file(self.my_output_dump_file_path)
        ground_truth_dump_data = dump_utils.read_numpy_file(self.ground_truth_dump_file_path)
        self._check_shape_valid(my_output_dump_data, ground_truth_dump_data)

        result, error_msg = self.manager.compare(
            my_output_dump_data.flatten(), ground_truth_dump_data.flatten(),
            {'my_output_dump_file': self.my_output_dump_file_path,
             'ground_truth_dump_file': self.ground_truth_dump_file_path,
             'shape_type': utils.get_shape_type(my_output_dump_data.shape)})
        self._print_result(result, error_msg, save_result)

    def _print_result(self: any, result: list, error_msg: list, save_result: bool) -> None:
        header = self.manager.get_result_title()
        title = ''
        line = ''
        for (algorithm_index, value) in enumerate(result):
            while len(line) < len(title):
                line += ConstManager.SPACE
            while len(title) < len(line):
                title += ConstManager.SPACE
            title += str(header[algorithm_index]) + ConstManager.SPACE
            line += str(value) + ConstManager.SPACE
        log.print_info_log(title)
        log.print_info_log(line)
        if error_msg:
            log.print_info_log(str(error_msg))
        if save_result:
            content = ""
            for index, data in enumerate(result):
                content += "%s: %s\n" % (header[index], data)
            file_name = 'file_result_%s.txt' % time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
            path_type = path_check.PathType.File
            summary_file_path = os.path.join(self.output_path, file_name)
            ret = path_check.check_output_path_valid(summary_file_path, exist=False, path_type=path_type)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)
            FileUtils.save_file(summary_file_path, content)
            log.print_info_log('The file compare result have been written to "%r".' % summary_file_path)

    def _check_shape_valid(self: any, my_output_dump_data: any, ground_truth_dump_data: any) -> None:
        my_output_shape = my_output_dump_data.shape
        ground_truth_shape = ground_truth_dump_data.shape
        if my_output_shape != ground_truth_shape:
            log.print_error_log("My output shape %s in '%r' does not match the ground truth shape %s in '%r'."
                                % (my_output_shape, self.my_output_dump_file_path, ground_truth_shape,
                                   self.ground_truth_dump_file_path))
            raise CompareError(CompareError.MSACCUCMP_INVALID_SHAPE_ERROR)
